from collections import namedtuple
from contextlib import contextmanager
from datetime import (
    datetime,
    timedelta,
)
import httplib
from jenkins import JenkinsException
import json
import logging
import os
import urllib2

import yaml

from jobs import (
    BUILD_REVISION,
    parse_tags,
    PUBLISH_REVISION,
)
from utility import (
    date_str,
    ensure_dir,
    FileInUse,
    locked_open_rw,
    parse_iso_date,
    temp_dir,
)


__metaclass__ = type


BUILDING = 'building'
SUCCEEDED = 'succeeded'
FAILED = 'failed'
PENDING = 'pending'
TERMINAL_STATUS = (SUCCEEDED, FAILED)


BLESS = 'bless'
CURSE = 'curse'
FINAL_RESULTS = (BLESS, CURSE)


CloudHealthJob = namedtuple('CloudHealthJob',
                            ['name', 'substrate', 'last_completed'])


class StateFileInUse(Exception):
    """Raised when the lock for the state file is held by another process."""
    pass


def get_substrate_from_parsed(tags):
    substrates = tags.get('substrate', [])
    if len(substrates) == 0:
        return None
    elif len(substrates) > 1:
        raise ValueError('Too many substrates: {}'.format(
            ','.join(substrates)))
    return substrates[0]


class StateFile:
    """Provide programmatic access to the state data."""

    def __init__(self, data=None, fileobj=None, s3_storage=None):
        if data is None:
            data = {
                'versions': {},
                'versions_by_revision': {},
                'test_builds': {},
            }
        self._data = data
        self._fileobj = fileobj
        self.logger = logging.getLogger('cidirector')
        self.s3_storage = s3_storage

    @classmethod
    def from_file(cls, fileobj, s3_storage):
        """Return a StateFile for a specified File object."""
        data = yaml.safe_load(fileobj)
        fileobj.seek(0)
        return cls(data, fileobj, s3_storage)

    @classmethod
    @contextmanager
    def acquire(cls, s3_storage):
        """Context manager that provides a StateFile and file _object.

        An exclusive lock is taken on the file.  If the lock cannot be
        acquired, StateFileInUse is raised.
        """
        try:
            with locked_open_rw(StateFile.file_name()) as state_file_obj:
                yield cls.from_file(state_file_obj, s3_storage), state_file_obj
        except FileInUse:
            raise StateFileInUse('State file in use.')

    @staticmethod
    def file_name():
        """Return the filename of the current user's state file."""
        return os.path.join(os.environ['HOME'], '.config/ci-director-state')

    @property
    def current_version_number(self):
        """Return the build number of the current version."""
        return self._data.get('current_version')

    @property
    def _current_version(self):
        current_version = self._data.get('current_version')
        if current_version is None:
            return None
        return self._data['versions'][current_version]

    def start_revision(self, branch, revno, revision_id, release_number,
                       build_id):
        """Record that building a revision has started.

        :param branch: Location of the branch being built.
        :param revno: Revno of the branch being built.
        :param revision_id: Revision-id of the branch being built.
        :param release_number: Version number of the Juju release.
        :param build_id: Number of the build.
        """
        assert build_id not in self._data['versions']
        self._data['current_version'] = build_id
        self._data['versions'][build_id] = {
            'branch': branch,
            'revno': revno,
            'revision_id': revision_id,
            'release_number': release_number,
            'tests': {},
            'build': {},
            'started': date_str(datetime.utcnow()),
            'cloud-health': {'substrates': {}}
        }
        builds = self._data['versions_by_revision'].setdefault(revision_id, [])
        builds.append(build_id)
        self.get_build_revision_job().start_build(build_id)

    def finish_revision(self, now):
        info = self.get_current_version()
        info['finished'] = date_str(now)

    def get_current_version_summary(self):
        """Retrieve a summary of the current revision.

        :return: a dict of build_number, revno, branch, revision_id.  (Same
            meaning as in start_revision).
        """
        current_version = self._current_version
        result = dict((k, current_version[k]) for k in
                      ['revno', 'branch', 'revision_id', 'release_number'])
        result['build_number'] = self._data['current_version']
        revno = current_version.get('revno')
        if revno is None:
            result['short_revision'] = current_version['revision_id'][:8]
        else:
            result['short_revision'] = 'r%d' % revno
        return result

    def get_build_revision_job(self):
        """Return the StateFileJob for build-revision."""
        version_data = self._current_version
        if version_data is None:
            return None
        # Handle legacy entries-- new entries will always have this set by
        # start_revision()
        build_data = version_data.setdefault(
            'build', {'status': PENDING, 'building': True})
        return StateFileJob(BUILD_REVISION, build_data, logger=self.logger)

    def publication_job(self):
        """Return the publication StateFileJob for the current version.

        Return None if there isn't a current version.
        """
        version_data = self._current_version
        if version_data is None:
            return None
        for job in self.iter_grouped_jobs():
            if job.test_id == PUBLISH_REVISION:
                return job
        tags = (
            'subjects=release,feature function=publish-metadata '
            'arch=all series=all substrate=none')
        return self.get_grouped_job(
            PUBLISH_REVISION, group_name='release-tests', failure_threshold=2,
            tags=tags, prerequisite_jobs=[BUILD_REVISION])

    def get_grouped_job(self, job_id, group_name=None, failure_threshold=None,
                        tags=None, vote=True, prerequisite_jobs=[]):
        """Retrieve the StateFileJob for a given grouped job."""
        if self._current_version is None:
            return None
        if group_name is None:
            group = self._current_version
        else:
            group = self._current_version.setdefault(group_name, {})
        default = {
            'status': PENDING,
            'tags': tags,
            'failures': 0,
            'vote': vote,
            'prerequisite_jobs': prerequisite_jobs,
        }
        job_dict = group.setdefault(job_id, default)
        return StateFileJob(job_id, job_dict, failure_threshold, vote=vote,
                            logger=self.logger,
                            prerequisite_jobs=prerequisite_jobs)

    def iter_grouped_jobs(self):
        """Iterate through all group StateFileJobs for the current version.

        WARNING: This method yields the discovered job, some jobs maybe
        created in the future.
        """
        if 'current_version' not in self._data:
            return
        for job_id, job_dict in self._current_version.items():
            if job_id == 'build' or not isinstance(job_dict, dict):
                continue
            if 'failures' in job_dict:
                yield StateFileJob(
                    job_id, job_dict, None,
                    vote=job_dict.get('vote', True), logger=self.logger,
                    prerequisite_jobs=job_dict.get('prerequisite_jobs', []))
            else:
                group = job_dict
                for job_id, job_dict in group.items():
                    if isinstance(job_dict, dict) and 'failures' in job_dict:
                        yield StateFileJob(
                            job_id, job_dict, None,
                            vote=job_dict.get('vote', True),
                            logger=self.logger,
                            prerequisite_jobs=job_dict.get(
                                'prerequisite_jobs', []))

    def revision_prepared(self, revision_id):
        """Determine whether a revision has already been prepared.

        :param revision_id: The revision to consider
        :return: True or False
        """
        builds = self._data['versions_by_revision'].get(revision_id, [])
        return bool(len(builds) > 0)

    def write(self):
        """Write this state data to the supplied file."""
        self._fileobj.seek(0)
        yaml.safe_dump(self._data, self._fileobj)
        self._fileobj.truncate()

    def save_result(self, status, build):
        """"Record the bless/curse result for the current version.

        :param status: BLESS or CURSE
        :param build: The number of the build where this was recorded.
        """
        if status not in FINAL_RESULTS:
            raise ValueError('Illegal value: %s' % status)
        self._current_version['result'] = {'status': status, 'build': build}

    def get_result(self):
        """Retrieve the bless/curse result for the current version.

        :return: a tuple of (status, build), or (None, None) if no result was
            recorded.
        """
        result = self._current_version.get('result')
        if result is None:
            return None, None
        return result['status'], result['build']

    def get_complete_failure_info(self, revision_build):
        """Return the complete information about failures for a revision build.

        Only those failures that contribute to the build being cursed are
        reported, not -devel failures.
        """
        revision_data = self._data['versions'][revision_build]
        return revision_data.get('complete_failure_info')

    def set_complete_failure_info(self, revision_build, failure_info):
        """Store the complete information about failures for a revision build.

        Only those failures that contribute to the build being cursed are
        stored, not -devel failures.
        """
        revision_data = self._data['versions'][revision_build]
        revision_data['complete_failure_info'] = failure_info

    def get_current_version(self):
        return self._current_version

    def build_info_to_s3(self):
        dest_url = '%s/result.json' % self.s3_storage.archive_version_url(
            self.current_version_number)
        with temp_dir() as work_dir:
            file_path = os.path.join(work_dir, 'result.json')
            with open(file_path, 'w') as result:
                info = self.get_current_version()
                json.dump(info, result, indent=2, sort_keys=True,
                          separators=(',', ': '))
            self.s3_storage.run_s3cmd(['--no-progress', 'put', file_path,
                                       dest_url])

    def jobs_sorted_by_dependency(self):
        """Topologically sort jobs by dependency and yield them."""
        jobs = list(self.iter_grouped_jobs())
        brj = self.get_build_revision_job()
        if brj is not None:
            jobs.append(brj)
        return topo_sort(jobs)

    def update_blocked(self):
        """Update the blocked status of the given jobs."""
        topo_jobs = list(self.jobs_sorted_by_dependency())
        for job in topo_jobs:
            job.update_blocked(topo_jobs)

    def get_cloud_health(self, version):
        version = self._data['versions'][version]
        return version.get('cloud-health', {'substrates': {}})

    def iter_recent_cloud_health(self, version, substrate, since):
        """Iterate health checks that ended within an hour of reference."""
        substrates = self.get_cloud_health(version).get('substrates')
        substrate_data = substrates.get(substrate, {})
        for test_id, builds in substrate_data.items():
            for build_id, build_data in builds.items():
                build_start = parse_iso_date(build_data['timestamp'])
                duration = timedelta(seconds=build_data['duration'] / 1000.0)
                build_end = build_start + duration
                if build_end < since:
                    continue
                yield (test_id, build_id, build_data)

    def update_cloud_health(self, version, job, state, timestamp, duration):
        version = self._data['versions'][version]
        cloud_health = version.setdefault('cloud-health', {})
        substrates = cloud_health.setdefault('substrates', {})
        substrate = substrates.setdefault(job.substrate, {})
        time_str = date_str(datetime_from_timestamp(timestamp))
        substrate.setdefault(job.name, {})[job.last_completed] = {
            'status': state, 'timestamp': time_str, 'duration': duration}

    def is_last_build_healthy(self, version, build_info, substrate):
        timestamp = datetime_from_timestamp(build_info['timestamp'])
        timestamp -= timedelta(hours=1)
        result = self.get_substrate_status(version, [substrate], timestamp)
        return result.get(substrate)

    def get_substrate_status(self, version, substrates, since):
        """Determine the status of the listed substrates since a time.

        Returns a map of substrates to status.  If no tests for that substrate
        ended since the specified time, there will be no entry for the
        substrate.  Otherwise, if any tests failed, there will be an entry
        with value False.  Otherwise, the value will be True.
        :param version: The build revision.
        :param substrates: A list of substrates to check.
        :param since: Only builds that were running since this time should be
            considered.
        """
        results = {}
        for substrate in substrates:
            for name, number, check in self.iter_recent_cloud_health(
                    version, substrate, since):
                if check['status'] == SUCCEEDED:
                    results.setdefault(substrate, True)
                if check['status'] == FAILED:
                    results[substrate] = False
                    break
        return results


def datetime_from_timestamp(timestamp):
    return datetime.utcfromtimestamp(timestamp / 1000.0)


def topo_sort(jobs):
    """Sort so that each job appears after its prerequisites."""
    by_name = dict((job.test_id, job) for job in jobs)
    depends = dict(
        (job.test_id, set(job.prerequisite_jobs))
        for job in jobs)
    while depends:
        to_remove = set()
        for job_name, prerequisites in depends.items():
            if len(prerequisites) == 0:
                yield by_name[job_name]
                to_remove.add(job_name)
                del depends[job_name]
        if len(to_remove) == 0:
            raise ValueError('Cyclic dependency: {}'.format(depends))
        for prerequisites in depends.values():
            prerequisites -= to_remove


def find_blockers(cur_job, all_jobs):
    """Find all blockers for this job.

    cur_job is the job to assess.
    all_jobs is a topo_sort()ed list of all jobs.
    Returns a list of names of jobs that block this one.
    """
    status = cur_job.get_status()
    result = []
    if status in (FAILED, SUCCEEDED, BUILDING):
        return result
    if cur_job.test_id == BUILD_REVISION:
        return result
    prerequisites = set(cur_job.prerequisite_jobs)
    prerequisites.add(BUILD_REVISION)
    depending_on = [
        job for job in all_jobs if job.test_id in prerequisites]
    result = [
        job.test_id for job in depending_on
        if job.is_blocked() or job.get_status() == FAILED]
    return result


class StateFileJob:
    """Representation of a Job in the StateFile."""

    def __init__(self, test_id, data, failure_threshold=1, vote=True,
                 logger=None, prerequisite_jobs=None):
        """Constructor.

        :param test_id: The ID of the job.
        :param data: The StateFile dict representing the job.
        :param failure_threshold: The number of build failures that cause the
            job to be considered a failure.
        """
        self.test_id = test_id
        # The failure_threshold comes from the data, but when absent, the
        # value passed to init is used, or finally fall back to 1.
        threshold = failure_threshold or 1
        # SFJ is a light-wrapper around the data dict. _data must be a
        # reference to the data dict because it is just a dict in a larger
        # dict. This data dance ensures the data has consistent keys with
        # sane defaults.
        if prerequisite_jobs is None:
            prerequisite_jobs = []
        if (test_id != BUILD_REVISION and
                BUILD_REVISION not in prerequisite_jobs):
            prerequisite_jobs.append(BUILD_REVISION)
        base_data = {
            'failure-threshold': threshold,
            'failures': 0,
            'last_build': None,
            'status': PENDING,
            'vote': vote,
            'building': False,
            'prerequisite_jobs': prerequisite_jobs,
        }
        base_data.update(data)
        data.update(base_data)
        self._data = data
        self.logger = logger
        self.vote = vote

    def __repr__(self):
        return 'StateFileJob<%r, %r>' % (self.test_id, self.get_status())

    def _summary(self):
        # We cannot directly hash the lists "prerequisite_jobs" and "blocked".
        data = {}
        data.update(self._data)
        data['prerequisite_jobs'] = tuple(data['prerequisite_jobs'])
        if 'blocked' in data:
            data['blocked'] = tuple(data['blocked'])
        return (self.test_id,
                tuple(sorted((k, v) for k, v in data.items())),
                self.failure_threshold)

    def __hash__(self):
        return hash(self._summary())

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        return self._summary() == other._summary()

    @property
    def build_number(self):
        return self._data['last_build']

    @property
    def last_build_healthy(self):
        return self._data.get('last_build_healthy')

    @last_build_healthy.setter
    def last_build_healthy(self, value):
        if value not in (True, False, None):
            raise ValueError('Invalid value: {}'.format(value))
        self._data['last_build_healthy'] = value

    @property
    def voting_build(self):
        if self.last_build_healthy is not None and not self.last_build_healthy:
            return False
        return self.vote

    @property
    def failure_threshold(self):
        return self._data['failure-threshold']

    @property
    def prerequisite_jobs(self):
        return self._data['prerequisite_jobs']

    def get_parsed_tags(self):
        tags = self._data.get('tags', '')
        if tags is None:
            return {}
        return parse_tags(tags)

    def get_substrate(self):
        tags = self.get_parsed_tags()
        return get_substrate_from_parsed(tags)

    def start_build(self, build_id):
        """Record a build of a provider test.

        The initial status is BUILDING.
        :param build_id: The build number of the build.
        """
        self._data.update({
            'last_build': build_id,
            'last_build_healthy': None,
            'building': True,
        })

    def get_status(self):
        """Return the current status of a job.

        If the current build is still running, return BUILDING.

        If the failure count reaches the threshold, and the status is
        PENDING, return FAILED.  Otherwise, return the 'status' value.
        """
        if self._data['building']:
            return BUILDING
        status = self._data['status']
        if (status == PENDING and
                self._data['failures'] >= self.failure_threshold):
            return FAILED
        return status

    def update_from_jenkins(self, jenkins, current_version, state_file):
        """Update to the last build's result via Jenkins.

        :param jenkins: A jenkins.Jenkins instance
        """
        last_build = self._data['last_build']
        if last_build is None:
            return
        try:
            info = jenkins.get_build_info(self.test_id, last_build)
        except JenkinsException as err:
            if str(err).endswith('number[%s] does not exist' % last_build):
                # A build can really disappear, see bug 1365808.
                self.start_build(None)
                self._data['building'] = False
                if self.logger is not None:
                    self.logger.warn(
                        'Build %s of job %s is missing' % (
                            last_build, self.test_id))
                return
            else:
                self.logger.warn(
                    'call failed; jenkins.get_build_info({}, {})'.format(
                        self.test_id, last_build))
                raise
        self.last_build_healthy = state_file.is_last_build_healthy(
            current_version, info, self.get_substrate())
        result = info.get('result')
        if result is None:
            return
        old_status = self.get_status()
        self.update_from_build_result(result)
        if old_status == BUILDING and old_status != self.get_status():
            self.archive_results(
                jenkins, info['artifacts'], current_version, info['url'],
                last_build, state_file.s3_storage)

    def update_from_build_result(self, result):
        """Update a provider test from its build result.

        Only FAILURE causes the failure count to increment.
        SUCCESS changes the status to SUCCEEDED, and other results change the
        status to PENDING.
        """
        def handle_FAILURE():
            self._data.update({
                'failures': self._data['failures'] + 1,
            })

        def handle_ABORTED():
            pass

        def handle_SUCCESS():
            self._data['status'] = SUCCEEDED

        handle_UNSTABLE = handle_FAILURE
        handle_NOT_BUILT = handle_ABORTED
        try:
            handler = locals()['handle_' + result]
        except KeyError:
            raise ValueError('No such status: %s' % result)
        self._data['building'] = False
        return handler()

    def archive_results(self, jenkins, artifacts, current_version, build_url,
                        build_number, s3_storage):
        """Copy artifacts of this job to an S3 bucket."""
        with temp_dir() as work_dir:
            dest_url = '%s/%s/build-%s/' % (
                s3_storage.archive_version_url(current_version), self.test_id,
                build_number)
            downloads = []
            for artifact in artifacts:
                display_path = artifact['displayPath']
                relative_path = artifact['relativePath']
                if display_path is None:
                    display_path = relative_path.partition("artifacts/")[-1]
                if display_path == 'empty':
                    continue
                source_url = '%s/artifact/%s' % (build_url, relative_path)
                extra_dirs = os.path.dirname(display_path)
                if extra_dirs:
                    ensure_dir(os.path.join(work_dir, extra_dirs))
                local_path = os.path.join(work_dir, display_path)
                downloads.append((local_path, source_url))
            local_path = os.path.join(work_dir, 'consoleText')
            source_url = '%s/%s' % (build_url, 'consoleText')
            downloads.append((local_path, source_url))
            for local_path, source_url in downloads:
                self.jenkins_download(jenkins, local_path, source_url)
            s3_storage.run_s3cmd(['--no-progress', 'sync', '%s/' % work_dir,
                                 dest_url])

    def jenkins_download(self, jenkins, local_path, source_url, size=4096):
        """Download a url from jenkins to a local file."""
        request = urllib2.Request(source_url)
        if jenkins.auth:
            request.add_header("Authorization", jenkins.auth)
        try:
            source_file = urllib2.urlopen(request)
            try:
                with open(local_path, "wb") as local_file:
                    while True:
                        data = source_file.read(size)
                        if not data:
                            break
                        local_file.write(data)
            finally:
                source_file.close()
        except (httplib.HTTPException, IOError) as e:
            if self.logger is not None:
                self.logger.info(
                    "failed to download jenkins url %s: %s",
                    source_url, e)
            raise

    def update_blocked(self, all_jobs):
        self._data['blocked'] = find_blockers(self, all_jobs)

    def is_blocked(self):
        """Return True if this job is blocked, else return False.

        A job is blocked if prerequisite jobs failed or are themselves
        blocked.
        """
        return len(self._data.get('blocked', [])) > 0
