from datetime import (
    datetime,
    timedelta,
    )
import logging
import os.path
from StringIO import StringIO
import subprocess
from textwrap import dedent
import urllib2
from unittest import TestCase

from mock import (
    call,
    Mock,
    patch,
)
import yaml

from cidirector.jobs import (
    BadTagsError,
    BUILD_REVISION,
    JENKINS_BUILD_RESULTS,
    PUBLISH_REVISION,
    SUCCESS,
)
from cidirector.storage import (
    BLESS,
    BUILDING,
    CloudHealthJob,
    CURSE,
    FAILED,
    find_blockers,
    PENDING,
    StateFile,
    StateFileJob,
    SUCCEEDED,
    topo_sort,
)
from cidirector.tests.test_cidirector import FakeJenkins
from cidirector.tests.utility import (
    finally_fail_job,
)
from cidirector.utility import (
    date_str,
    parse_iso_date,
    S3Storage,
    temp_dir,
)


__metaclass__ = type


def make_build_info(offset=0):
    # Make a timestamp that is exactly one hour after the conventional
    # CloudHealthJob ended, plus an optional offset.
    timestamp = 123000 + 321000 + 3600000 + offset
    return {'timestamp': timestamp}


class FakeJob:

    def __init__(self, test_id, status=SUCCEEDED, blocked=False,
                 prerequisite_jobs=frozenset()):
        self.test_id = test_id
        self._status = status
        self._blocked = blocked
        self.prerequisite_jobs = prerequisite_jobs

    def get_status(self):
        return self._status

    def is_blocked(self):
        return self._blocked


class TestTopoSort(TestCase):

    def test_topo_sort(self):
        jobs = [
            FakeJob('foo', prerequisite_jobs={'baz'}),
            FakeJob('bar'),
            FakeJob('baz', prerequisite_jobs={'qux'}),
            FakeJob('qux')]
        sorted_names = [j.test_id for j in topo_sort(jobs)]
        self.assertTrue(sorted_names.index('baz') < sorted_names.index('foo'))
        self.assertTrue(sorted_names.index('qux') < sorted_names.index('baz'))


class TestFindBlockers(TestCase):

    def test_build_revision_blockage(self):
        build_revision = FakeJob(BUILD_REVISION, FAILED)
        foo_job = FakeJob('foo', PENDING)
        result = find_blockers(foo_job, [build_revision, foo_job])
        self.assertEqual([BUILD_REVISION], result)

    def test_build_revision_no_blockage(self):
        build_revision = FakeJob(BUILD_REVISION, SUCCEEDED)
        foo_job = FakeJob('foo', PENDING)
        result = find_blockers(foo_job, [build_revision, foo_job])
        self.assertEqual([], result)

    def cur_job_status(self, status, blocked):
        build_revision = FakeJob(BUILD_REVISION, FAILED)
        foo_job = FakeJob('foo', status)
        result = find_blockers(foo_job, [build_revision, foo_job])
        if blocked:
            self.assertEqual([BUILD_REVISION], result)
        else:
            self.assertEqual([], result)

    def test_cur_job_status(self):
        self.cur_job_status(BUILDING, False)
        self.cur_job_status(SUCCEEDED, False)
        self.cur_job_status(FAILED, False)
        self.cur_job_status(PENDING, True)

    def test_prerequisites(self):
        foo_job = FakeJob('foo', FAILED)
        bar_job = FakeJob('bar', PENDING, prerequisite_jobs={'foo'})
        self.assertEqual(['foo'], find_blockers(bar_job, [foo_job, bar_job]))
        bar_job.prerequisite_jobs = []
        self.assertEqual([], find_blockers(bar_job, [foo_job, bar_job]))


class TestStateFileJob(TestCase):

    def test_get_parsed_tags(self):
        sfj = StateFileJob('foo', {
            'tags': ('subjects=substrate,cpc function=upgrade substrate=aws'
                     ' arch=amd64 series=precise')})
        self.assertEqual(sfj.get_parsed_tags(), {
            'arch': ['amd64'],
            'function': ['upgrade'],
            'series': ['precise'],
            'subjects': ['substrate', 'cpc'],
            'substrate': ['aws'],
            })

    def test_get_parsed_tags_invalid(self):
        sfj = StateFileJob('foo', {'tags': 'foo'})
        with self.assertRaises(BadTagsError):
            sfj.get_parsed_tags()

    def test_get_parsed_tags_None(self):
        sfj = StateFileJob('foo', {'tags': None})
        self.assertEqual(sfj.get_parsed_tags(), {})

    def test_get_substrate(self):
        sfj = StateFileJob('foo', {'tags': 'substrate=foo'})
        self.assertEqual(sfj.get_substrate(), 'foo')

    def test_get_substrate_no_tags(self):
        sfj = StateFileJob('foo', {})
        self.assertIs(sfj.get_substrate(), None)

    def test_get_substrate_multiple(self):
        sfj = StateFileJob('foo', {'tags': 'substrate=foo,bar'})
        with self.assertRaisesRegexp(
                ValueError, 'Too many substrates: foo,bar'):
            sfj.get_substrate()

    def test_start_build(self):
        sfj = StateFileJob('foo', {})
        sfj.start_build(59)
        self.assertEqual(
            sfj._data, {
                'last_build': 59, 'last_build_healthy': None,
                'status': PENDING, 'failures': 0, 'failure-threshold': 1,
                'vote': True, 'building': True,
                'prerequisite_jobs': ['build-revision']})

    def test_start_build_keeps_failures(self):
        sfj = StateFileJob(
            'foo', {'last_build': 49, 'status': PENDING, 'failures': 1})
        sfj.start_build(59)
        self.assertEqual(
            sfj._data, {
                'last_build': 59, 'last_build_healthy': None, 'status':
                PENDING, 'failures': 1, 'failure-threshold': 1, 'vote': True,
                'building': True, 'prerequisite_jobs': ['build-revision']})

    def start_build_keeps_success(self):
        sfj = StateFileJob(
            'foo',
            {'last_build': 49, 'status': SUCCEEDED, 'failures': 1,
             'building': False})
        sfj.start_build(50)
        self.assertEqual(
            sfj._data, {
                'last_build': 50, 'status': SUCCEEDED, 'failures': 1,
                'failure-threshold': 1, 'vote': True, 'building': True})

    def test_start_build_resets_last_build_healthy(self):
        sfj = StateFileJob('foo', {'last_build_healthy': True})
        self.assertEqual(sfj.last_build_healthy, True)
        sfj.start_build(60)
        self.assertIs(sfj.last_build_healthy, None)

    def test_update_from_build_result_failed(self, status='FAILURE'):
        sfj = StateFileJob('foo', {})
        sfj.start_build(59)
        sfj.update_from_build_result(status)
        self.assertEqual(
            sfj._data,
            {'last_build': 59, 'last_build_healthy': None, 'status': PENDING,
             'failures': 1, 'failure-threshold': 1, 'vote': True, 'building':
             False, 'prerequisite_jobs': ['build-revision']})

    def test_update_from_build_result_unstable(self):
        return self.test_update_from_build_result_failed('UNSTABLE')

    def test_update_from_build_result_until_test_failed(self,
                                                        status='FAILURE'):
        sfj = StateFileJob('foo', {})
        sfj.start_build(55)
        while sfj.get_status() != FAILED:
            sfj.update_from_build_result(status)
        status = sfj._data['status']
        self.assertEqual(status, PENDING)

    def test_update_from_build_result_until_test_unstable(self):
        return self.test_update_from_build_result_until_test_failed('UNSTABLE')

    def test_update_from_build_result_accepts_only_valid(self):
        sfj = StateFileJob('foo', {})
        for value in JENKINS_BUILD_RESULTS:
            sfj.start_build(3)
            sfj.update_from_build_result(value)
        with self.assertRaises(ValueError):
            sfj.update_from_build_result('INVALID')

    def test_update_from_build_result_aborted(self, status='ABORTED'):
        sfj = StateFileJob('foo', {})
        for value in JENKINS_BUILD_RESULTS:
            sfj.start_build(3)
            sfj.update_from_build_result(status)
        self.assertEqual(sfj._data,
                         {'status': PENDING, 'failures': 0, 'last_build': 3,
                          'last_build_healthy': None, 'failure-threshold': 1,
                          'vote': True, 'building': False,
                          'prerequisite_jobs': ['build-revision']})

    def test_update_from_build_result_not_built(self):
        self.test_update_from_build_result_aborted('NOT_BUILT')

    def test_get_status(self):
        sfj = StateFileJob('foo', {}, 5)
        self.assertIs(sfj.get_status(), PENDING)
        sfj.start_build(55)
        self.assertEqual(sfj.get_status(), BUILDING)
        sfj.update_from_build_result('FAILURE')
        self.assertEqual(sfj.get_status(), PENDING)
        finally_fail_job(self, sfj)
        self.assertEqual(sfj.get_status(), FAILED)

    def test_get_status_build_after_failure(self):
        sfj = StateFileJob('foo', {'failures': 0}, 5)
        finally_fail_job(self, sfj)
        sfj.start_build(6)
        self.assertEqual(sfj.get_status(), BUILDING)

    def test_get_status_extra_build_after_success(self):
        sfj = StateFileJob(
            'foo',
            {'failures': 0, 'last_build': 63, 'status': SUCCEEDED,
             'building': False},
            5)
        sfj.start_build(64)
        self.assertEqual(sfj.get_status(), BUILDING)
        sfj.update_from_build_result('FAILURE')
        # The result SUCCEDED is not overwritten by the new failure.
        self.assertEqual(sfj.get_status(), SUCCEEDED)

    def test_init_with_failure_threshold(self):
        data = {'status': PENDING, 'failures': 0, 'failure-threshold': 3}
        # The failure_threshold can come from data.
        data['failure-threshold'] = 3
        sfj = StateFileJob('foo', data, None)
        self.assertEqual(3, sfj.failure_threshold)
        self.assertEqual(3, sfj._data['failure-threshold'])
        self.assertEqual(0, sfj._data['failures'])
        # The failure_threshold in the data takes precedence over the arg
        # passed to init.
        sfj = StateFileJob('foo', data, 5)
        self.assertEqual(3, sfj.failure_threshold)
        self.assertEqual(3, sfj._data['failure-threshold'])
        self.assertEqual(0, sfj._data['failures'])
        # When neither the data doesn't care, the value comes from the arg
        del data['failure-threshold']
        sfj = StateFileJob('foo', data, 5)
        self.assertEqual(5, sfj.failure_threshold)
        self.assertEqual(5, sfj._data['failure-threshold'])
        self.assertEqual(0, sfj._data['failures'])
        # When neither the data and the failure_threshold arg doesn't care,
        # the value is set to 1
        del data['failure-threshold']
        sfj = StateFileJob('foo', data, None)
        self.assertEqual(1, sfj.failure_threshold)
        self.assertEqual(1, sfj._data['failure-threshold'])

    def test_update_from_jenkins(self):
        jenkins = FakeJenkins({
            'foo': {
                'build_number': 2,
                'building': False,
                'result': SUCCESS,
                },
        })
        sfj = StateFileJob('foo', {'last_build': 2}, None)
        sf = StateFile()
        sf._data['versions'][12] = {}
        sfj.update_from_jenkins(jenkins, 12, sf)
        self.assertEqual(SUCCEEDED, sfj._data['status'])

    def test_update_from_jenkins_adapts_specialised_jobs(self):
        jenkins = FakeJenkins({
            PUBLISH_REVISION: {
                'build_number': 2, 'building': False, 'result': 'SUCCESS'},
        })
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20.1', 2)
        for method in [sf.publication_job]:
            sfj = sf.publication_job()
            sfj._data['last_build'] = 2
            sfj.update_from_jenkins(jenkins, 2, sf)
            self.assertEqual('succeeded', sfj._data['status'])

    def test_update_from_jenkins_triggers_archiving_artifacts(self):
        build_number = 1234
        foo_job = 'foo'
        foo_data = {
            'status': PENDING, 'failures': 0, 'failure-threshold': 3,
            'build_number': build_number, 'last_build': build_number,
            'building': True,
        }
        jobs = {foo_job: foo_data}
        current_version = 4567
        jenkins = FakeJenkins(jobs, current_version)
        sf = StateFile()
        sf._data['versions'][4567] = {}
        sfj = StateFileJob(foo_job, foo_data, None)
        with patch.object(sfj, 'archive_results') as mocked_archive_results:
            self.assertEqual(BUILDING, sfj.get_status())
            sfj.update_from_jenkins(jenkins, current_version, sf)
            self.assertEqual(BUILDING, sfj.get_status())
            self.assertEqual(0, len(mocked_archive_results.mock_calls))

            artifacts = [
                {
                    'fileName': 'some-product.tar.gz',
                    'relativePath': 'build/some-product.tar.gz',
                },
            ]
            jenkins.set_build_result(
                foo_job, build_number, 'SUCCESS', artifacts)
            sfj.update_from_jenkins(jenkins, current_version, sf)
            self.assertEqual(SUCCEEDED, sfj.get_status())
            mocked_archive_results.assert_called_once_with(
                jenkins, artifacts, current_version,
                'fakettp://fake.fake/job/foo/1234', build_number, None)

    def test_update_from_jenkins_missing_build(self):
        build_number = 1234
        foo_job = 'foo'
        foo_data = {
            'status': PENDING, 'failures': 0, 'failure-threshold': 3,
            'build_number': build_number, 'last_build': build_number - 1,
            'last_build_healthy': True, 'building': True,
        }
        jobs = {foo_job: foo_data}
        current_version = 4567
        jenkins = FakeJenkins(jobs, current_version)
        logger = logging.getLogger()
        handler = logging.handlers.MemoryHandler(10)
        logger.addHandler(handler)
        sfj = StateFileJob(foo_job, foo_data, None, logger=logger)
        sfj.update_from_jenkins(jenkins, current_version, StateFile())
        self.assertEqual(PENDING, sfj._data['status'])
        self.assertIs(None, sfj._data['last_build'])
        self.assertIs(None, sfj._data['last_build_healthy'])
        self.assertFalse(sfj._data['building'])
        self.assertEqual(
            'Build 1233 of job foo is missing', handler.buffer[0].msg)

    def get_update_from_jenkins_params(self, substrate):
        build_number = 1234
        foo_job = 'foo'
        foo_data = {
            'status': PENDING, 'failures': 0, 'failure-threshold': 3,
            'build_number': build_number, 'last_build': build_number,
            'last_build_healthy': None, 'building': True,
        }
        if substrate:
            foo_data['tags'] = 'substrate={}'.format(substrate)
        jobs = {foo_job: foo_data}
        current_version = 4567
        jenkins = FakeJenkins(jobs, current_version)
        sf = StateFile()
        sf.start_revision('branch', 1, 'rev-id', '1.22.37', current_version)
        sfj = StateFileJob(foo_job, foo_data, None)
        return jenkins, sf, sfj

    def test_update_from_jenkins_last_build_healthy_no_substrate(self):
        jenkins, sf, sfj = self.get_update_from_jenkins_params(None)
        sfj.update_from_jenkins(jenkins, 4567, sf)
        self.assertIs(sfj.last_build_healthy, None)

    def test_update_from_jenkins_last_build_healthy_succeeded(self):
        jenkins, sf, sfj = self.get_update_from_jenkins_params('bws')
        chjob = CloudHealthJob(substrate='bws', name='bar', last_completed=0)
        sfj.update_from_jenkins(jenkins, 4567, sf)
        self.assertIs(sfj.last_build_healthy, None)
        sf.update_cloud_health(4567, chjob, SUCCEEDED, 0, 3600000)
        sfj.update_from_jenkins(jenkins, 4567, sf)
        self.assertIs(sfj.last_build_healthy, True)

    def test_update_from_jenkins_last_build_healthy_failed(self):
        jenkins, sf, sfj = self.get_update_from_jenkins_params('bws')
        chjob = CloudHealthJob(substrate='bws', name='bar', last_completed=0)
        sf.update_cloud_health(4567, chjob, FAILED, 1, 3600000)
        sfj.update_from_jenkins(jenkins, 4567, sf)
        self.assertIs(sfj.last_build_healthy, False)

    def test_archive_results(self):
        sfj = StateFileJob('foo', {})
        artifacts = [
            {
                'fileName': 'build.log',
                'displayPath': 'build.log',
                'relativePath': 'logs/build.log',
            },
            {
                'fileName': 'bar',
                'displayPath': 'bar',
                'relativePath': 'bar',
            },
            {
                'fileName': 'empty',
                'displayPath': 'empty',
                'relativePath': 'empty',
            },
        ]
        fake_jenkins = FakeJenkins({})
        with patch.object(
                sfj, 'jenkins_download', autospec=True) as mocked_download:
            with patch.object(
                    subprocess, 'check_output') as mocked_check_output:
                sfj.archive_results(
                    fake_jenkins, artifacts, 1234,
                    'http://ci.testing/job/foo/42', 4567,
                    S3Storage('foo', 'bar'))
        call_params = mocked_download.call_args_list
        self.assertEqual(fake_jenkins, call_params[0][0][0])
        self.assertEqual(
            'http://ci.testing/job/foo/42/artifact/logs/build.log',
            call_params[0][0][2])
        self.assertEqual(fake_jenkins, call_params[1][0][0])
        self.assertEqual(
            'http://ci.testing/job/foo/42/artifact/bar',
            call_params[1][0][2])
        self.assertEqual(fake_jenkins, call_params[2][0][0])
        self.assertEqual(
            'http://ci.testing/job/foo/42/consoleText',
            call_params[2][0][2])
        local_dir = mocked_check_output.call_args_list[0][0][0][5]
        self.assertEqual('/', local_dir[-1])
        expected = ([
            's3cmd', '-c', 'foo', '--no-progress',
            'sync', local_dir,
            'bar/juju-ci/products/version-1234/foo/build-4567/',
        ])
        mocked_check_output.assert_called_once_with(expected)
        self.assertEqual(3, len(call_params))

    def test_archive_results_subdirs(self):
        artifacts = [
            {
                'fileName': 'cloud-init-output.log.gz',
                'displayPath': '0/cloud-init-output.log.gz',
                'relativePath': 'artifacts/0/cloud-init-output.log.gz',
            },
            {
                'fileName': 'cloud-init-output.log.gz',
                'displayPath': '1/cloud-init-output.log.gz',
                'relativePath': 'artifacts/1/cloud-init-output.log.gz',
            },
        ]
        self._archive_results_test(artifacts)

    def test_archive_results_subdirs_only_relative(self):
        artifacts = [
            {
                'displayPath': None,
                'relativePath': 'artifacts/0/cloud-init-output.log.gz',
            },
            {
                'displayPath': None,
                'relativePath': 'artifacts/1/cloud-init-output.log.gz',
            },
        ]
        self._archive_results_test(artifacts)

    def _archive_results_test(self, artifacts):
        sfj = StateFileJob('foo', {})
        fake_jenkins = FakeJenkins({})
        with patch.object(
                sfj, 'jenkins_download', autospec=True) as mocked_download:
            with patch.object(
                    subprocess, 'check_output') as mocked_check_output:
                with patch('cidirector.storage.ensure_dir',
                           autospec=True) as mocked_ensure_dir:
                    sfj.archive_results(
                        fake_jenkins, artifacts, 1234,
                        'http://ci.testing/job/foo/42', 4567,
                        S3Storage('foo', 'bar'))
        local_dir = mocked_check_output.call_args_list[0][0][0][5]
        self.assertEqual('/', local_dir[-1])
        self.assertEqual(mocked_ensure_dir.mock_calls, [
            call(os.path.join(local_dir, '0')),
            call(os.path.join(local_dir, '1')),
        ])
        self.assertEqual(mocked_download.mock_calls, [
            call(fake_jenkins,
                 os.path.join(local_dir, '0/cloud-init-output.log.gz'),
                 'http://ci.testing/job/foo/42/artifact/artifacts/0/'
                 'cloud-init-output.log.gz'),
            call(fake_jenkins,
                 os.path.join(local_dir, '1/cloud-init-output.log.gz'),
                 'http://ci.testing/job/foo/42/artifact/artifacts/1/'
                 'cloud-init-output.log.gz'),
            call(fake_jenkins, os.path.join(local_dir, 'consoleText'),
                 'http://ci.testing/job/foo/42/consoleText'),
        ])
        expected = ([
            's3cmd', '-c', 'foo', '--no-progress',
            'sync', local_dir,
            'bar/juju-ci/products/version-1234/foo/build-4567/',
        ])
        mocked_check_output.assert_called_once_with(expected)

    def test_is_blocked(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20.1', 2)
        foo_job = sf.get_grouped_job('foo')
        bar_job = sf.get_grouped_job('bar', prerequisite_jobs=['foo'])
        baz_job = sf.get_grouped_job('baz', prerequisite_jobs=['bar'])

        sf.update_blocked()
        self.assertFalse(foo_job.is_blocked())
        self.assertFalse(bar_job.is_blocked())

        finally_fail_job(self, foo_job)
        sf.update_blocked()
        self.assertTrue(bar_job.is_blocked())
        self.assertTrue(baz_job.is_blocked())

    def test_voting_build_vote(self):
        sfj = StateFileJob('foo', {})
        self.assertEqual(sfj.voting_build, True)
        sfj.vote = False
        self.assertEqual(sfj.voting_build, False)
        sfj.last_build_healthy = False
        self.assertEqual(sfj.voting_build, False)

    def test_voting_build_vote_last_build_healthy(self):
        sfj = StateFileJob('foo', {})
        sfj.last_build_healthy = True
        self.assertEqual(sfj.voting_build, True)
        sfj.last_build_healthy = None
        self.assertEqual(sfj.voting_build, True)
        sfj.last_build_healthy = False
        self.assertEqual(sfj.voting_build, False)

    def test_jenkins_download(self):
        sfj = StateFileJob('test', {})
        source_url = "http://ci.testing"
        expected_content = "some file contents"
        expected_filelike = StringIO(expected_content)
        auth_value = "preformed auth string"

        class MockJenkins:

            auth = auth_value

        with temp_dir() as directory:
            local_path = os.path.join(directory, "anArtifact")
            with patch.object(urllib2, "urlopen", autospec=True,
                              return_value=expected_filelike) as mock_urlopen:
                sfj.jenkins_download(MockJenkins(), local_path, source_url)
            self.assertEqual(mock_urlopen.call_count, 1)
            first_arg = mock_urlopen.call_args[0][0]
            self.assertIsInstance(first_arg, urllib2.Request)
            self.assertEqual(first_arg.get_full_url(), source_url)
            self.assertEqual(first_arg.get_header("Authorization"), auth_value)
            with open(local_path, "rb") as f:
                self.assertEqual(f.read(), expected_content)

    def test_jenkins_download_handles_ioerror(self):
        mock_logger = Mock(spec=logging.Logger)
        source_url = "http://ci.testing"
        error = IOError("fail")
        sfj = StateFileJob('test', {}, logger=mock_logger)

        class MockJenkins:

            auth = "an auth string"

        with temp_dir() as directory:
            local_path = os.path.join(directory, "anArtifact")
            with patch.object(urllib2, "urlopen", autospec=True,
                              side_effect=error):
                with self.assertRaises(IOError) as cm:
                    sfj.jenkins_download(MockJenkins(), local_path, source_url)
            self.assertIs(cm.exception, error)
            self.assertFalse(os.path.exists(local_path))

        mock_logger.info.assert_called_once_with(
            "failed to download jenkins url %s: %s", source_url, error
        )


class TestStateFile(TestCase):

    def test_file_name(self):
        with patch.dict(os.environ, {'HOME': '/home/jrandomx'}):
            self.assertEqual('/home/jrandomx/.config/ci-director-state',
                             StateFile.file_name())

    def test_acquire_no_file(self):
        with temp_dir() as directory:
            filename = os.path.join(directory, 'state-file')
            with patch.object(StateFile, 'file_name', return_value=filename):
                with StateFile.acquire(None):
                    self.assertTrue(os.path.exists(filename))

    def test_acquire_retains_contents(self):
        with temp_dir() as directory:
            filename = os.path.join(directory, 'state-file')
            with open(filename, 'w') as f:
                f.write('foo')
            with patch.object(StateFile, 'file_name', return_value=filename):
                with StateFile.acquire(S3Storage('foo', 'bar')) as (sf, f):
                    f.seek(0)
                    self.assertEqual(f.read(), 'foo')

    def test_from_file(self):
        data = ['foo', 'bar', 'baz']
        fileobj = StringIO()
        yaml.safe_dump(data, fileobj)
        fileobj.seek(0)
        s3_storage = S3Storage('qux', 'quxx')
        sf = StateFile.from_file(fileobj, s3_storage)
        self.assertEqual(data, sf._data)
        self.assertIs(fileobj, sf._fileobj)
        self.assertIs(sf.s3_storage, sf.s3_storage)
        sf2 = StateFile.from_file(fileobj, s3_storage)
        self.assertEqual(data, sf2._data)

    def test_write_seeks_truncates(self):
        fileobj = StringIO()
        yaml.dump('foo', fileobj)
        foo_yaml = fileobj.getvalue()
        fileobj.write('trailing')
        sf = StateFile('foo', fileobj)
        sf.write()
        self.assertEqual(foo_yaml, fileobj.getvalue())

    def test_start_revision(self):
        sf = StateFile()
        self.assertEqual(sf._data, {'versions': {}, 'test_builds': {},
                                    'versions_by_revision': {}})
        before = datetime.utcnow()
        sf.start_revision('lp:foo', 25, 'revid-blah', '14.6.3', 57)
        after = datetime.utcnow()
        started = parse_iso_date(sf._data['versions'][57].pop('started'))
        self.assertTrue(before <= started <= after)
        self.assertEqual(sf._data, {
            'current_version': 57,
            'versions': {
                57: {
                    'branch': 'lp:foo',
                    'revno': 25,
                    'revision_id': 'revid-blah',
                    'build': {
                        'failures': 0,
                        'last_build': 57,
                        'last_build_healthy': None,
                        'status': PENDING,
                        'failure-threshold': 1,
                        'vote': True,
                        'building': True,
                        'prerequisite_jobs': [],
                    },
                    'tests': {},
                    'release_number': '14.6.3',
                    'cloud-health': {'substrates': {}}
                }
            },
            'test_builds': {},
            'versions_by_revision': {'revid-blah': [57]},
        })

    def test_finish_revision(self):
        sf = StateFile()
        sf.start_revision('a', 5, 'b', '0.21', 'c')
        self.assertIs(None, sf.get_current_version().get('finished'))
        sf.finish_revision(datetime(1, 2, 3, 4, 5, 6, 7))
        self.assertEqual(sf.get_current_version()['finished'],
                         '0001-02-03T04:05:06.000007Z')

    def test_get_current_version_summary(self):
        sf = StateFile()
        sf.start_revision('lp:foo', 25, 'revid-blah', '1.6', 57)
        self.assertEqual(sf.get_current_version_summary(), {
            'branch': 'lp:foo',
            'revno': 25,
            'short_revision': 'r25',
            'revision_id': 'revid-blah',
            'build_number': 57,
            'release_number': '1.6',
        })

    def test_last_build_healthy_missing(self):
        sfj = StateFileJob('foo-id', {})
        self.assertIs(sfj.last_build_healthy, None)

    def test_last_build_healthy_value(self):
        sfj = StateFileJob('foo-id', {'last_build_healthy': True})
        self.assertEqual(sfj.last_build_healthy, True)

    def test_last_build_healthy_invalid(self):
        sfj = StateFileJob('foo-id', {'last_build_healthy': 'asdf'})
        self.assertEqual(sfj.last_build_healthy, 'asdf')

    def test_set_last_build_healthy(self):
        sfj = StateFileJob('foo-id', {})
        sfj.last_build_healthy = True
        self.assertEqual(sfj._data['last_build_healthy'], True)
        sfj.last_build_healthy = False
        self.assertEqual(sfj._data['last_build_healthy'], False)
        sfj.last_build_healthy = None
        self.assertIs(sfj._data['last_build_healthy'], None)

    def test_set_last_build_invalid(self):
        sfj = StateFileJob('foo-id', {})
        with self.assertRaisesRegexp(ValueError, 'Invalid value: foo'):
            sfj.last_build_healthy = 'foo'

    def test_get_current_version_summary_no_revno(self):
        sf = StateFile()
        sf.start_revision('lp:foo', None, '1234567890', '1.6', 57)
        self.assertEqual(sf.get_current_version_summary(), {
            'branch': 'lp:foo',
            'revno': None,
            'short_revision': '12345678',
            'revision_id': '1234567890',
            'build_number': 57,
            'release_number': '1.6',
        })

    def test_revision_prepared(self):
        sf = StateFile()
        self.assertIs(False, sf.revision_prepared('ab'))
        self.assertIs(False, sf.revision_prepared('cd'))
        sf.start_revision('a', 5, 'ab', '1.6', '6')
        self.assertIs(True, sf.revision_prepared('ab'))
        self.assertIs(False, sf.revision_prepared('cd'))

    def test_roundtrip(self):
        sf = StateFile(fileobj=StringIO())
        sf.start_revision('a', 5, 'ab', '1.6', '6')
        sf.write()
        sf._fileobj.seek(0)
        sf2 = StateFile.from_file(sf._fileobj, S3Storage('foo', 'bar'))
        self.assertEqual(sf._data, sf2._data)

    def test_save_result_bless(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        sf.save_result(BLESS, 90)
        self.assertEqual(
            sf._current_version['result'], {'status': BLESS, 'build': 90})

    def test_save_result_curse(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        sf.save_result(CURSE, 90)
        self.assertEqual(
            sf._current_version['result'], {'status': CURSE, 'build': 90})

    def test_save_result_pending(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        with self.assertRaisesRegexp(ValueError, 'Illegal value: pending'):
            sf.save_result(PENDING, 90)

    def test_save_result_other(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        with self.assertRaisesRegexp(ValueError, 'Illegal value: other'):
            sf.save_result('other', 90)

    def test_get_result(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        self.assertEqual((None, None), sf.get_result())
        sf.save_result(CURSE, 90)
        self.assertEqual((CURSE, 90), sf.get_result())
        sf.save_result(BLESS, 123)
        self.assertEqual((BLESS, 123), sf.get_result())

    def test_publication_job(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        self.assertNotIn(PUBLISH_REVISION, sf._data['versions'][2])
        pubjob = sf.publication_job()
        self.assertIn(
            PUBLISH_REVISION,
            sf._data['versions'][2]['release-tests'])
        self.assertIs(PENDING, pubjob.get_status())
        stored = sf._data['versions'][2]['release-tests'][PUBLISH_REVISION]
        self.assertEqual(['build-revision'], stored['prerequisite_jobs'])

    def test_publication_job_is_resourceful(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20.1', 2)
        self.assertNotIn(PUBLISH_REVISION, sf._data['versions'][2])
        job = sf.get_grouped_job(
            PUBLISH_REVISION, 'release-tests', {'marker': True})
        pub_job = sf.publication_job()
        self.assertEqual(job, pub_job)

    def test_publication_job_no_version(self):
        sf = StateFile()
        self.assertIs(None, sf.publication_job())

    def test_get_build_revision_job(self):
        sf = StateFile()
        self.assertIs(None, sf.get_build_revision_job())
        sf.start_revision('a', 1, 'b', '1.6', 2)
        self.assertEqual(
            sf.get_build_revision_job(),
            StateFileJob('build-revision', {
                'status': PENDING,
                'failure-threshold': 1,
                'failures': 0,
                'last_build': 2,
                'last_build_healthy': None,
                'building': True,
            }))

    def test_get_grouped_job(self):
        # Calling get_grouped_job() without a group creates a job in
        # the anonymous top-level group with a failure_threshold of 1.
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20', 7)
        grouped_job = sf.get_grouped_job('foo')
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': None, 'failures': 0,
                     'vote': True, 'failure-threshold': 1,
                     'last_build': None, 'building': False,
                     'prerequisite_jobs': ['build-revision']}),
            sf._data['versions'][7].items())
        self.assertEqual(grouped_job.failure_threshold, 1)
        finally_fail_job(self, grouped_job)
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': None, 'failures': 1,
                     'vote': True, 'failure-threshold': 1,
                     'last_build': None, 'building': False,
                     'prerequisite_jobs': ['build-revision']}),
            sf._data['versions'][7].items())

    def test_get_grouped_job_with_group_with_failure_threshold(self):
        # Calling get_grouped_job() with a group creates both the group
        # if needed and the job. The failure_threshold is set.
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20', 7)
        grouped_job = sf.get_grouped_job('foo', 'baz', 2)
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': None, 'failures': 0,
                     'vote': True, 'failure-threshold': 2,
                     'last_build': None, 'building': False,
                     'prerequisite_jobs': ['build-revision']}),
            sf._data['versions'][7]['baz'].items())
        self.assertEqual(grouped_job.failure_threshold, 2)
        finally_fail_job(self, grouped_job)
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': None, 'failures': 2,
                     'vote': True, 'failure-threshold': 2,
                     'last_build': None, 'building': False,
                     'prerequisite_jobs': ['build-revision']}),
            sf._data['versions'][7]['baz'].items())

    def test_get_grouped_job_with_tags(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20', 7)
        sf.get_grouped_job('foo', 'baz', 2, 'qux')
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': 'qux', 'failures': 0,
                     'vote': True, 'failure-threshold': 2,
                     'last_build': None, 'building': False,
                     'prerequisite_jobs': ['build-revision']}),
            sf._data['versions'][7]['baz'].items())

    def test_get_grouped_job_with_prereqisite_jobs(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20', 7)
        sf.get_grouped_job('foo', 'baz', prerequisite_jobs=['bar'])
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': None, 'failures': 0,
                     'vote': True, 'failure-threshold': 1,
                     'last_build': None, 'building': False,
                     'prerequisite_jobs': ['bar', 'build-revision']}),
            sf._data['versions'][7]['baz'].items())

    def test_iter_grouped_jobs(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20,0', 7)
        self.assertItemsEqual([], sf.iter_grouped_jobs())
        sf.get_grouped_job('foo', 'baz', 2)
        sf.get_grouped_job('bar', None, 2)
        sf.save_result('curse', 7)
        data = {
            'status': PENDING, 'tags': None, 'failures': 0, 'vote': True,
            'failure-threshold': 2}
        self.assertItemsEqual([
            StateFileJob('foo', data, failure_threshold=None),
            StateFileJob('bar', data, failure_threshold=None),
        ], sf.iter_grouped_jobs())

    def test_get_set_complete_failure_info(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        info = sf.get_complete_failure_info(2)
        self.assertIs(None, info)
        sf.set_complete_failure_info(2, {'foo-test': 1, 'bar-test': 2})
        info = sf.get_complete_failure_info(2)
        self.assertEqual({'foo-test': 1, 'bar-test': 2}, info)

    def test_build_info_to_s3(self):
        s3cmd_call_data = []

        def catch_temp_file_contents(call_params):
            file_path = call_params[5]
            s3cmd_call_data.append(file_path)
            s3cmd_call_data.append(yaml.load(open(file_path)))

        sf = StateFile(s3_storage=S3Storage('foo', 'bar'))
        sf.start_revision('a', 1, 'b', '1.6', 2)
        with patch.object(
                subprocess, 'check_output',
                side_effect=catch_temp_file_contents) as check_output_mock:
            sf.build_info_to_s3()
        check_output_mock.assert_called_once_with([
            's3cmd', '-c', 'foo', '--no-progress',
            'put', s3cmd_call_data[0],
            'bar/juju-ci/products/version-2/result.json'
        ])
        expected_build_result = sf.get_current_version()
        self.assertEqual(expected_build_result, s3cmd_call_data[1])

    def test_build_info_to_s3_pretty(self):
        file_contents = []

        def catch_temp_file_contents(call_params):
            file_path = call_params[5]
            file_contents.append(open(file_path).read())

        sf = StateFile(s3_storage=S3Storage('foo', 'bar'))
        sf.start_revision('a', 1, 'b', '1.6', 2)
        sf._data['versions'][2]['started'] = date_str(datetime(1, 2, 3))
        with patch.object(subprocess, 'check_output',
                          side_effect=catch_temp_file_contents):
            sf.build_info_to_s3()
        expected_contents = dedent("""\
            {
              "branch": "a",
              "build": {
                "building": true,
                "failure-threshold": 1,
                "failures": 0,
                "last_build": 2,
                "last_build_healthy": null,
                "prerequisite_jobs": [],
                "status": "pending",
                "vote": true
              },
              "cloud-health": {
                "substrates": {}
              },
              "release_number": "1.6",
              "revision_id": "b",
              "revno": 1,
              "started": "0001-02-03T00:00:00Z",
              "tests": {}
            }""")
        self.assertEqual(expected_contents, file_contents[0])

    def test_jobs_sorted_by_dependency(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20.1', 2)
        job_1 = sf.get_grouped_job('job_1', prerequisite_jobs=['job_4'])
        job_2 = sf.get_grouped_job('job_2', prerequisite_jobs=['job_1'])
        job_3 = sf.get_grouped_job(
            'job_3', prerequisite_jobs=['job_2', 'job_1'])
        job_4 = sf.get_grouped_job(
            'job_4', prerequisite_jobs=['build-revision'])
        self.assertEqual(
            [sf.get_build_revision_job(), job_4, job_1, job_2, job_3],
            list(sf.jobs_sorted_by_dependency()))

    def test_sort_jobs_by_dependency_cyclic_dependency(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20.1', 2)
        sf.get_grouped_job('job_1', prerequisite_jobs=['job_2'])
        sf.get_grouped_job('job_2', prerequisite_jobs=['job_1'])

        def test_func():
            return list(sf.jobs_sorted_by_dependency())

        self.assertRaises(ValueError, test_func)

    def test_get_cloud_health_empty(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        self.assertEqual(sf.get_cloud_health(25), {'substrates': {}})

    def test_get_cloud_health(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        sf.update_cloud_health(25, CloudHealthJob('foo', 'bar', 'baz'),
                               SUCCEEDED, 123456789000, 987654321)
        self.assertEqual(sf.get_cloud_health(25), {'substrates': {
            'bar': {'foo': {'baz': {
                'status': SUCCEEDED,
                'timestamp': '1973-11-29T21:33:09Z',
                'duration': 987654321,
                }}}
        }})

    def test_update_cloud_health(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        sf.update_cloud_health(25, CloudHealthJob('foo', 'bar', 'baz'),
                               FAILED, 123000, 321)
        self.assertEqual(sf._data['versions'][25]['cloud-health'],
                         {'substrates': {'bar': {'foo': {'baz': {
                             'status': FAILED,
                             'timestamp': '1970-01-01T00:02:03Z',
                             'duration': 321,
                             }}}}})

    def get_statefile(self, duration):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        self.assertEqual(list(sf.iter_recent_cloud_health(
            25, 'foo', datetime(2000, 1, 1))), [])
        sf.update_cloud_health(25, CloudHealthJob('foo', 'bar', 28),
                               FAILED, 123000, duration)
        return sf

    def test_iter_recent_cloud_health_too_old(self):
        sf = self.get_statefile(0)
        since = datetime(2000, 1, 1)
        self.assertEqual(list(sf.iter_recent_cloud_health(
            25, 'bar', since)), [])

    def test_iter_recent_cloud_heath_since_start(self):
        sf = self.get_statefile(0)
        since = datetime.utcfromtimestamp(123)
        expected = ('foo', 28, {
            'status': FAILED, 'timestamp': '1970-01-01T00:02:03Z',
            'duration': 0
            })
        self.assertEqual(
            list(sf.iter_recent_cloud_health(25, 'bar', since)), [expected])
        self.assertEqual(
            list(sf.iter_recent_cloud_health(25, 'bar', since +
                 timedelta(seconds=1))), [])

    def test_iter_recent_cloud_heath_since_end(self):
        sf = self.get_statefile(321000)
        since = datetime.utcfromtimestamp(123 + 321)
        expected = ('foo', 28, {
            'status': FAILED, 'timestamp': '1970-01-01T00:02:03Z',
            'duration': 321000
            })
        self.assertEqual(
            list(sf.iter_recent_cloud_health(25, 'bar', since)), [expected])

    def test_is_last_build_healthy_no_substrate(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        self.assertIs(sf.is_last_build_healthy(
            25, make_build_info(), None), None)

    def test_is_last_build_healthy_no_result(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        self.assertIs(sf.is_last_build_healthy(
            25, make_build_info(), 'bar'), None)

    def test_is_last_build_healthy_true(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        sf.update_cloud_health(25, CloudHealthJob('foo', 'bar', 28),
                               SUCCEEDED, 123000, 321000)
        self.assertEqual(sf.is_last_build_healthy(
            25, make_build_info(), 'bar'), True)

    def test_is_last_build_healthy_false(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        sf.update_cloud_health(25, CloudHealthJob('foo', 'bar', 28),
                               FAILED, 123000, 321000)
        self.assertEqual(sf.is_last_build_healthy(
            25, make_build_info(), 'bar'), False)

    def test_is_last_build_healthy_no_recent(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        sf.update_cloud_health(25, CloudHealthJob('foo', 'bar', 28),
                               SUCCEEDED, 123000, 321000)
        self.assertIs(sf.is_last_build_healthy(
            25, make_build_info(offset=1), 'bar'), None)

    def test_get_substrate_status_no_data(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        initial_substrates = ['foo', 'bar', 'baz']
        status = sf.get_substrate_status(
            25, initial_substrates, datetime.utcfromtimestamp(0))
        self.assertEqual(status, {})

    def test_get_substrate_status_some_unhealthy(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        sf.update_cloud_health(25, CloudHealthJob('foo', 'sub1', 28),
                               FAILED, 123000, 321000)
        sf.update_cloud_health(25, CloudHealthJob('foo', 'sub3', 28),
                               FAILED, 123000, 321000)
        status = sf.get_substrate_status(
            25, ['sub1', 'sub2', 'sub3', 'sub4'],
            datetime.utcfromtimestamp(0))
        self.assertEqual(status, {'sub1': False, 'sub3': False})

    def test_get_substrate_status_old_unhealthy(self):
        sf = StateFile()
        sf.start_revision('branch', 38, 'rev-id', '1.2.3', 25)
        sf.update_cloud_health(25, CloudHealthJob('foo', 'sub1', 28),
                               FAILED, 123000, 321000)
        sf.update_cloud_health(25, CloudHealthJob('foo', 'sub3', 28),
                               FAILED, 123000, 322000)
        status = sf.get_substrate_status(
            25, ['sub1', 'sub2', 'sub3', 'sub4'],
            datetime.utcfromtimestamp(444))
        self.assertEqual(status, {'sub1': False, 'sub3': False})
        status = sf.get_substrate_status(
            25, ['sub1', 'sub2', 'sub3', 'sub4'],
            datetime.utcfromtimestamp(445))
        self.assertEqual(status, {'sub3': False})
