from calendar import timegm
from contextlib import contextmanager
from copy import deepcopy
from datetime import (
    date,
    datetime,
    timedelta,
    )
import json
import os
from unittest import TestCase

from cidirector.jobs import (
    BUILD_REVISION,
    FAILURE,
    SUCCESS,
    )
from cidirector.storage import (
    BLESS,
    BUILDING,
    CURSE,
    FAILED,
    PENDING,
    SUCCEEDED,
    )
from cidirector.update_outcome import (
    BuildFiles,
    find_cloud_health_builds,
    find_last_build_sickly,
    get_health_match_data,
    get_revision_build,
    iter_job_builds,
    load_cloud_health,
    NotLocked,
    OutcomeJob,
    OutcomeJobSource,
    OutcomeState,
    scan_new_builds,
    )
from cidirector.utility import temp_dir
from cidirector.tests.test_cidirector import (
    FakeJenkins,
    make_build_info,
    )


def make_jenkins1():
    jenkins = FakeJenkins({})
    info = make_build_info(1, 27, result='SUCCESS')
    info3 = make_build_info(3, 27, result='SUCCESS')
    jenkins.set_build_info('steve', info)
    jenkins.set_build_info('steve', info3)
    jenkins.job_info['steve']['description'] = (
        '[ci-director]\nfailure_threshold: 2')
    return jenkins, info, info3


def add_cloud_builds(jenkins, timestamp):
    jenkins_timestamp = datetime_to_timestamp(timestamp)
    info1 = make_build_info(1, 27, result='SUCCESS')
    info1['timestamp'] = jenkins_timestamp
    info2 = make_build_info(2, 27, result='SUCCESS')
    info2['timestamp'] = jenkins_timestamp
    jenkins.set_build_info('dave', info1)
    jenkins.set_build_info('dave', info2)
    jenkins.job_info['dave']['description'] = (
        '[cloud-health]\n'
        'failure_threshold: 2')
    return info1, info2


class TestIterJobBuilds(TestCase):

    def test_iter_job_builds(self):
        jenkins = make_jenkins1()[0]
        steve_info = jenkins.get_job_info('steve')
        self.assertItemsEqual([(steve_info, [1, 3])],
                              iter_job_builds(jenkins, {}))

    def test_skip_seen(self):
        jenkins = make_jenkins1()[0]
        steve_info = jenkins.get_job_info('steve')
        self.assertItemsEqual([(steve_info, [3])],
                              iter_job_builds(jenkins, {'steve': 1}))
        self.assertItemsEqual([(steve_info, [])],
                              iter_job_builds(jenkins, {'steve': 3}))

    def test_skip_no_config(self):
        jenkins = make_jenkins1()[0]
        jenkins.job_info['steve']['description'] = '#[ci-director]'
        self.assertItemsEqual([], iter_job_builds(jenkins, {'steve': 1}))

    def test_skip_random(self):
        jenkins = make_jenkins1()[0]
        jenkins.job_info['steve']['description'] = '[random]'
        self.assertItemsEqual([], iter_job_builds(jenkins, {'steve': 1}))

    def test_include_cloud_health(self):
        jenkins = make_jenkins1()[0]
        jenkins.job_info['steve']['description'] = '[cloud-health]'
        steve_info = jenkins.get_job_info('steve')
        self.assertItemsEqual([(steve_info, [1, 3])],
                              iter_job_builds(jenkins, {}))

    def test_include_build_revision(self):
        jenkins = FakeJenkins({})
        info = make_build_info(1, 27, result='SUCCESS')
        jenkins.set_build_info(BUILD_REVISION, info)
        br_info = jenkins.get_job_info(BUILD_REVISION)
        self.assertItemsEqual([(br_info, [1])], iter_job_builds(jenkins, {}))


def datetime_to_timestamp(dt):
    return timegm(dt.timetuple()) * 1000


class TestGetRevisionBuild(TestCase):

    def test_get_revision_build(self):
        jenkins, info, info3 = make_jenkins1()
        self.assertEqual(27, get_revision_build('steve', info))
        self.assertEqual(27, get_revision_build('steve', info3))

    def test_build_revision_no_explict_revision_build(self):
        jenkins = FakeJenkins({})
        info = make_build_info(4, result='SUCCESS')
        jenkins.set_build_info('build-revision', info)
        self.assertEqual(4, get_revision_build('build-revision', info))


class TestScanNewBuilds(TestCase):

    def test_no_data(self):
        jenkins = FakeJenkins({})
        self.assertEqual(({}, {}, {}), scan_new_builds({}, jenkins))

    def test_new_data(self):
        jenkins, info, info3 = make_jenkins1()
        expected = (
            {27: {'steve': {
                'builds': {'1': info, '3': info3},
                'config': {'failure_threshold': 2},
                }}},
            {}, {'steve': 3})
        self.assertEqual(expected, scan_new_builds({}, jenkins))

    def test_new_data_build_revision(self):
        jenkins = FakeJenkins({})
        info = make_build_info(1, 27, result='SUCCESS')
        jenkins.set_build_info(BUILD_REVISION, info)
        expected = (
            {1: {BUILD_REVISION: {
                'builds': {'1': info},
                'config': None,
                }}},
            {}, {BUILD_REVISION: 1})
        self.assertEqual(expected, scan_new_builds({}, jenkins))

    def test_new_data_cloud_health(self):
        jenkins, info, info3 = make_jenkins1()
        info['timestamp'] = datetime_to_timestamp(datetime(2015, 11, 26))
        info3['timestamp'] = datetime_to_timestamp(datetime(2015, 11, 26))
        jenkins.job_info['steve']['description'] = (
            '[cloud-health]\n'
            'failure_threshold: 2')
        expected = (
            {},
            {'2015-11-26': {'steve': {
                'builds': {'1': info, '3': info3},
                'config': {'failure_threshold': 2},
                }}}, {'steve': 3})
        result = scan_new_builds({}, jenkins)
        self.assertEqual(expected, result)

    def test_new_data_mixed(self):
        jenkins, info, info3 = make_jenkins1()
        c_info1, c_info2 = add_cloud_builds(
            jenkins, datetime(2015, 11, 26))
        expected = (
            {27: {'steve': {
                'builds': {'1': info, '3': info3},
                'config': {'failure_threshold': 2},
                }}},
            {'2015-11-26': {'dave': {
                'builds': {'1': c_info1, '2': c_info2},
                'config': {'failure_threshold': 2},
                }}}, {'steve': 3, 'dave': 2})
        result = scan_new_builds({}, jenkins)
        self.assertEqual(expected, result)

    def test_no_last_completed(self):
        jenkins, info, info3 = make_jenkins1()
        job_info = jenkins.get_job_info('steve')
        job_info['lastCompletedBuild'] = None
        expected = (
            {27: {'steve': {
                'builds': {'1': info, '3': info3},
                'config': {'failure_threshold': 2},
                }}},
            {}, {})
        self.assertEqual(expected, scan_new_builds({}, jenkins))


class FakeOutcomeState:

    def __init__(self):
        self.files = {}

    def load_json(self, filename):
        try:
            content = self.files[filename]
        except KeyError:
            return None
        else:
            return deepcopy(content)

    def write_json(self, filename, content):
        self.files[filename] = deepcopy(content)

    def outcome_filename(self, filename):
        return filename


class TestBuildFiles(TestCase):

    def test_for_revision_builds(self):
        state = object()
        build_files = BuildFiles.for_revision_builds(state)
        self.assertEqual(build_files.outcome_state, state)
        self.assertEqual(build_files.filename_template,
                         'revision-build-{}.json')

    def test_for_cloud_health(self):
        state = object()
        build_files = BuildFiles.for_cloud_health(state)
        self.assertEqual(build_files.outcome_state, state)
        self.assertEqual(build_files.filename_template,
                         'cloud-health-{}.json')

    def test_filename(self):
        bf = BuildFiles(None, 'foo-{}-bar')
        self.assertEqual('foo-baz-bar', bf.filename('baz'))

    def test_update_revision_builds_no_existing(self):
        state = FakeOutcomeState()
        build_files = BuildFiles.for_revision_builds(state)
        list(build_files.update_files(
            {27: {'foo': {
                'builds': {'23': {}},
                'config': {'tags': 'foo=bar'},
                }}}))
        self.assertEqual(
            {'foo': {
                'builds': {'23': {}},
                'config': {'tags': 'foo=bar'},
                }},
            state.load_json('revision-build-27.json'))

    def test_update_revision_builds_existing(self):
        state = FakeOutcomeState()
        state.write_json('revision-build-27.json', {'foo': {
            'builds': {'23': {}},
            'config': {'tags': 'foo=bar'},
                }})
        build_files = BuildFiles.for_revision_builds(state)
        list(build_files.update_files(
            {27: {'foo': {
                'builds': {'24': {}},
                'config': {'tags': 'foo=baz'},
                }}}
            ))
        self.assertEqual({'foo': {
            'builds': {'23': {}, '24': {}},
            'config': {'tags': 'foo=baz'},
            }}, state.load_json('revision-build-27.json'))

    def test_merge_revision_build_data_config_no_build(self):
        orig = {'foo': {
            'builds': {'23': {}},
            'config': {'tags': 'foo=bar'},
            }}
        existing = deepcopy(orig)
        new = {'foo': {'config': {}, 'builds': {}}}
        BuildFiles.merge_revision_build_data(existing, new)
        self.assertEqual(existing['foo']['config'], orig['foo']['config'])

    def test_merge_revision_build_data_config_with_build(self):
        orig = {'foo': {
            'builds': {'23': {}},
            'config': {'tags': 'foo=bar'},
            }}
        existing = deepcopy(orig)
        new = {'foo': {'config': {}, 'builds': {'1': {}}}}
        BuildFiles.merge_revision_build_data(existing, new)
        self.assertEqual(existing['foo']['config'], new['foo']['config'])


class TestLoadCloudHealth(TestCase):

    def test_load_cloud_health(self):
        orig_health = {
            '2014-01-02': {'foo': {'builds': {23: {}}}},
            '2014-01-03': {'bar': {'builds': {24: {}}}},
            '2014-01-04': {
                'bar': {'builds': {25: {}}},
                'foo': {'builds': {26: {}}},
                },
        }
        state = FakeOutcomeState()
        cloud_health_builds = BuildFiles.for_cloud_health(state)
        list(cloud_health_builds.update_files(orig_health))
        result = load_cloud_health(
            cloud_health_builds, [date(2014, 1, 2), date(2014, 1, 4)])
        self.assertEqual({
            '2014-01-02': {'foo': {'builds': {23: {}}}},
            '2014-01-04': {
                'bar': {'builds': {25: {}}},
                'foo': {'builds': {26: {}}},
                },
            }, result)

    def test_load_cloud_health_no_result(self):
        # Ensure 2014-01-02 is missing, rather than assigned to None, in the
        # output.
        orig_health = {
            '2014-01-03': {'bar': {'builds': {24: {}}}},
            '2014-01-04': {
                'bar': {'builds': {25: {}}},
                'foo': {'builds': {26: {}}},
                },
        }
        state = FakeOutcomeState()
        cloud_health_builds = BuildFiles.for_cloud_health(state)
        list(cloud_health_builds.update_files(orig_health))
        result = load_cloud_health(
            cloud_health_builds, [date(2014, 1, 2), date(2014, 1, 4)])
        self.assertEqual({
            '2014-01-04': {
                'bar': {'builds': {25: {}}},
                'foo': {'builds': {26: {}}},
                },
            }, result)


def timestamp_from_datetime(supplied):
    since_epoch = (supplied - datetime(1970, 1, 1, 0, 0)).total_seconds()
    return since_epoch * 1000


class TestFindCloudHealthBuilds(TestCase):

    default_datetime = datetime(2014, 5, 6, 12, 30)

    default_timestamp = timestamp_from_datetime(default_datetime)

    @classmethod
    def make_cloud_health_builds(cls):
        return {'2014-05-06': {
            'foo-bar': {
                'config': {'tags': 'substrate=baz'},
                'builds': {'123': {
                    'timestamp': cls.default_timestamp,
                    'duration': 30000,
                    }}},
            }}

    def test_find_cloud_health_builds_substrate(self):
        cloud_health_builds = self.make_cloud_health_builds()
        self.assertItemsEqual([('foo-bar', {
            'timestamp': self.default_timestamp,
            'duration': 30000,
            })], find_cloud_health_builds(
                cloud_health_builds, 'baz', self.default_datetime,
                self.default_datetime))
        self.assertItemsEqual([], find_cloud_health_builds(
            cloud_health_builds, 'qux', self.default_datetime,
            self.default_datetime))

    def test_find_cloud_health_builds_timestamp_start(self):
        cloud_health_builds = self.make_cloud_health_builds()
        start = datetime(2014, 5, 6, 12, 30)
        end = datetime(2014, 5, 6, 14, 30)
        expected_build = ('foo-bar', {
            'timestamp': self.default_timestamp,
            'duration': 30000,
            })
        self.assertItemsEqual([expected_build], find_cloud_health_builds(
            cloud_health_builds, 'baz', start + timedelta(seconds=30), end))
        self.assertItemsEqual([], find_cloud_health_builds(
            cloud_health_builds, 'baz', start + timedelta(seconds=31), end))

    def test_find_cloud_health_builds_timestamp_end(self):
        cloud_health_builds = self.make_cloud_health_builds()
        start = self.default_datetime - timedelta(seconds=30)
        end = self.default_datetime
        expected_build = ('foo-bar', {
            'timestamp': self.default_timestamp,
            'duration': 30000,
            })
        self.assertItemsEqual([expected_build], find_cloud_health_builds(
            cloud_health_builds, 'baz', start, end))
        self.assertItemsEqual([], find_cloud_health_builds(
            cloud_health_builds, 'baz', start, end - timedelta(seconds=1)))

    def test_find_cloud_health_builds_no_duration(self):
        cloud_health_builds = self.make_cloud_health_builds()
        builds = cloud_health_builds['2014-05-06']['foo-bar']['builds']
        del builds['123']['duration']
        start = self.default_datetime - timedelta(seconds=30)
        end = self.default_datetime
        self.assertItemsEqual([], find_cloud_health_builds(
            cloud_health_builds, 'baz', start, end))


class TestFindLastBuildSickly(TestCase):

    def test_find_last_build_sickly_none(self):
        last_build_sickly = find_last_build_sickly({}, {})
        self.assertEqual(({}, {}), last_build_sickly)

    def make_data(self, substrate='baz', offset=None, result=FAILURE):
        if offset is None:
            offset = timedelta()
        chb = TestFindCloudHealthBuilds.make_cloud_health_builds()
        build = chb['2014-05-06']['foo-bar']['builds']['123']
        build.update({
            'result': result,
            'number': 123,
            })
        start = TestFindCloudHealthBuilds.default_datetime + offset
        health_match = {'3059': {'foo': {
            'start': start, 'end': start, 'substrate': substrate,
            }}}
        return health_match, chb

    def test_find_last_build_sickly_match(self):
        health_match, chb = self.make_data()
        last_build_sickly = find_last_build_sickly(health_match, chb)
        self.assertEqual(({'3059': set(['foo'])}, {
            'baz': {'foo-bar': {'123': {'status': 'failed'}}}
            }), last_build_sickly)

    def test_find_last_build_wrong_substrate(self):
        health_match, chb = self.make_data(substrate='qux')
        last_build_sickly = find_last_build_sickly(health_match, chb)
        self.assertEqual(({'3059': set()}, {}), last_build_sickly)

    def test_find_last_build_wrong_time(self):
        health_match, chb = self.make_data(offset=timedelta(hours=2))
        last_build_sickly = find_last_build_sickly(health_match, chb)
        self.assertEqual(({'3059': set()}, {}), last_build_sickly)

    def test_find_last_build_success(self):
        health_match, chb = self.make_data(result=SUCCESS)
        last_build_sickly = find_last_build_sickly(health_match, chb)
        self.assertEqual(({'3059': set()}, {
            'baz': {'foo-bar': {'123': {'status': 'succeeded'}}}
            }), last_build_sickly)


class TestOutcomeJob(TestCase):

    def test_from_json_default(self):
        job = OutcomeJob.from_json('foo', {
            'config': {},
            'builds': {},
            }, True)
        self.assertEqual(PENDING, job.status)
        self.assertIs(None, job.build_number)
        self.assertEqual(0, job.failure_count)
        self.assertIs(True, job.voting_build)
        self.assertEqual(None, job.tags)
        self.assertEqual(set(), job.prerequisite_jobs)
        self.assertEqual([], job.blocked)

    def test_from_json_building(self):
        job = OutcomeJob.from_json('foo', {
            'config': {},
            'builds': {'1': {'building': True, 'number': 1}},
            }, True)
        self.assertEqual(BUILDING, job.status)

    def test_from_json_build_number(self):
        job = OutcomeJob.from_json('foo', {
            'config': {},
            'builds': {
                '1': {'building': True, 'number': 1},
                '2': {'building': True, 'number': 2},
                '5': {'building': True, 'number': 5},
                },
            }, True)
        self.assertEqual(5, job.build_number)

    def test_from_json_succeeded(self):
        job = OutcomeJob.from_json('foo', {
            'config': {},
            'builds': {'1': {
                'building': True,
                'number': 1,
                'result': SUCCESS,
                }},
            }, True)
        self.assertEqual(SUCCEEDED, job.status)

    def test_from_json_failed(self):
        job = OutcomeJob.from_json('foo', {
            'config': {},
            'builds': {'1': {
                'building': False,
                'number': 1,
                'result': FAILURE,
                }},
            }, True)
        self.assertEqual(FAILED, job.status)

    def is_blocked(self):
        job = OutcomeJob(
            'foo', build_number=5, status=SUCCESS, failure_count=27,
            voting_build=False, tags='', prerequisite_jobs=[], blocked=[])
        self.assertIs(False, job.is_blocked())
        job.blocked.append('q')
        self.assertIs(True, job.is_blocked())


class TestOutcomeJobSource(TestCase):

    def test_from_build_data(self):
        build_data = {
            'foo': {'config': {}, 'builds': {}},
            'bar': {'config': {}, 'builds': {}},
            }
        expected = [OutcomeJob.from_json(k, v, True)
                    for k, v in build_data.items()]
        source = OutcomeJobSource.from_build_data(build_data, set())
        self.assertEqual(expected, source.jobs)

    def test_from_build_data_sets_blocked(self):
        build_data = {
                'foo': {'config': {'requires': 'bar'}, 'builds': {}},
                'bar': {'config': {}, 'builds': {
                    '1': {'result': FAILURE, 'number': 1, 'building': False},
                    }},
            }
        source = OutcomeJobSource.from_build_data(
                build_data, {'foo': True, 'bar': True})
        by_name = dict((j.test_id, j) for j in source.jobs)
        self.assertEqual(['bar'], by_name['foo'].blocked)
        self.assertEqual([], by_name['bar'].blocked)

    def test_get_candidate_determine_result_jobs(self):
        jobs = []
        source = OutcomeJobSource(jobs)
        self.assertIs(jobs, source.get_candidate_determine_result_jobs())


class TestGetHealthMatchData(TestCase):

    def test_get_health_match_data(self):
        start = datetime(2014, 5, 6, 12, 30)
        start_timestamp = timestamp_from_datetime(start)
        by_revision_build = {'3059': {'foo': {
            'config': {'tags': 'substrate=bar'},
            'builds': {
                '2': {'timestamp': 0, 'number': 2},
                '11': {
                    'timestamp': start_timestamp, 'duration': 3600000,
                    'number': 11},
                },
            },
            }}
        self.assertEqual(({'3059': {
            'foo': {
                'substrate': 'bar',
                'start': start - timedelta(hours=1),
                'end': start + timedelta(hours=1),
                },
            }}, set([start.date()])), get_health_match_data(by_revision_build))

    def test_get_health_match_data_no_config(self):
        start = datetime(2014, 5, 6, 12, 30)
        start_timestamp = timestamp_from_datetime(start)
        by_revision_build = {'3059': {'foo': {
            'config': None,
            'builds': {
                '2': {'timestamp': 0, 'number': 2},
                '11': {
                    'timestamp': start_timestamp, 'duration': 3600000,
                    'number': 11},
                },
            },
            }}
        self.assertEqual(({'3059': {}}, set()),
                         get_health_match_data(by_revision_build))

    def test_get_health_match_data_no_builds(self):
        by_revision_build = {'3059': {'foo': {
            'config': {'tags': 'substrate=bar'},
            'builds': {},
            }}}
        self.assertEqual(
            ({'3059': {}}, set()), get_health_match_data(by_revision_build))


class TestOutcomeState(TestCase):

    @staticmethod
    @contextmanager
    def make_outcome_state():
        with temp_dir() as workspace:
            yield OutcomeState(workspace)

    def test_dir_creation(self):
        with temp_dir() as parent_dir:
            workspace = os.path.join(parent_dir, 'workspace')
            state = OutcomeState(workspace)
            self.assertFalse(os.path.lexists(workspace))
            with state:
                self.assertTrue(os.path.isdir(workspace))

    def test_initial_values(self):
        with temp_dir() as workspace:
            state = OutcomeState(workspace)
            self.assertEqual(workspace, state.outcome_dir)
            self.assertIs(None, state.last_completed_file)
            self.assertIs(None, state.data)

    def test_outcome_filename(self):
        with self.make_outcome_state() as state:
            self.assertEqual(os.path.join(state.outcome_dir, 'foo'),
                             state.outcome_filename('foo'))

    def test_as_context_manager(self):
        with self.make_outcome_state() as state:
            self.assertIs(None, state.last_completed_file)
            with state:
                self.assertIsNot(None, state.last_completed_file)
            self.assertIs(None, state.last_completed_file)

    def test_write_json(self):
        data = {'foo': 'bar'}
        with self.make_outcome_state() as state:
            filename = state.outcome_filename('foo')
            with self.assertRaises(NotLocked):
                state.write_json(filename, data)
            self.assertFalse(os.path.exists(filename))
            with state:
                state.write_json(filename, data)
                with open(filename) as f:
                    self.assertEqual('{\n  "foo": "bar"\n}\n', f.read())
            with self.assertRaises(NotLocked):
                state.write_json(filename, data)

    @staticmethod
    def get_revision_build_data():
        return {
            'build-revision': {'builds': {'32': {
                'number': 32,
                'building': False,
                'result': SUCCESS,
                'artifacts': {},
                'actions': [{
                    'parameters': [
                        {
                            'name': 'branch',
                            'value': 'gitbranch:master:github.com/juju/juju',
                            },
                        {
                            'name': 'revision',
                            'value': 'asdf',
                            },
                        ],
                    }],
                'timestamp': 100,
                'duration': 200,
                }}}
            }

    @staticmethod
    def get_expected_mongo_data():
        return {
            'tests': {'build-revision': {
                    'status': SUCCEEDED,
                    'name': 'build-revision',
                    'tags': ('subjects=release function=make-tarball arch=all'
                             ' series=all substrate=none'),
                    'last_build': 32,
                    'vote': True,
                    'failures': 0,
                    'blocked': [],
                }},
            'complete_failure_info': {}, 'release_number': None,
            'finished': '1970-01-01T00:00:00.300000Z',
            'result': {'status': BLESS},
            'branch': 'gitbranch:master:github.com/juju/juju',
            'revision_id': 'asdf',
            'build': {'last_build': 32},
            'cloud-health': {'substrates': {'aws': 'foo'}}
            }

    def test_make_mongo_data(self):
        mongo_data = OutcomeState.make_mongo_data(
                32, self.get_revision_build_data(), set(), {'aws': 'foo'})
        self.assertEqual(self.get_expected_mongo_data(), mongo_data)

    def test_make_mongo_data_release_number(self):
        rb_data = self.get_revision_build_data()
        rb_data['build-revision']['builds']['32']['artifacts'] = [{
                'fileName': 'juju-core_1.25-delta1.tar.gz',
            }]
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, set(), {})
        self.assertEqual('1.25-delta1', mongo_data['release_number'])

    def test_make_mongo_data_tags(self):
        rb_data = self.get_revision_build_data()
        rb_data['foo'] = {
                'builds': {},
                'config': {}
            }
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, set(), {})
        self.assertIs(None, mongo_data['tests']['foo']['tags'])
        rb_data['foo']['config']['tags'] = 'foo=bar'
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, set(), {})
        self.assertEqual('foo=bar', mongo_data['tests']['foo']['tags'])

    def test_make_mongo_data_vote(self):
        rb_data = self.get_revision_build_data()
        rb_data['foo'] = {
                'builds': {},
                'config': {}
            }
        lbs_data = set()
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, lbs_data, {})
        self.assertEqual(True, mongo_data['tests']['foo']['vote'])
        rb_data['foo']['config']['vote'] = False
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, lbs_data, {})
        self.assertEqual(False, mongo_data['tests']['foo']['vote'])

    @classmethod
    def get_job_json(cls, timestamp=100, duration=200):
        build_1 = {
            'number': 1,
            'building': False,
            'timestamp': timestamp,
            'duration': duration,
            }
        return {
                'builds': {'1': build_1},
                'config': {}
            }

    @classmethod
    def get_foo_pending(cls):
        rb_data = cls.get_revision_build_data()
        rb_data['foo'] = cls.get_job_json()
        return rb_data

    def test_make_mongo_data_job_status_pending(self):
        rb_data = self.get_foo_pending()
        lbs_data = set()
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, lbs_data, {})
        self.assertEqual(PENDING, mongo_data['result']['status'])

    def test_make_mongo_data_job_status_bless(self):
        rb_data = self.get_foo_pending()
        rb_data['foo']['builds']['1']['result'] = SUCCESS
        lbs_data = set()
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, lbs_data, {})
        self.assertEqual(BLESS, mongo_data['result']['status'])

    def test_make_mongo_data_job_status_curse(self):
        rb_data = self.get_foo_pending()
        rb_data['foo']['builds']['1']['result'] = FAILURE
        lbs_data = set()
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, lbs_data, {})
        self.assertEqual(CURSE, mongo_data['result']['status'])

    def test_make_mongo_data_blocked(self):
        rb_data = self.get_foo_pending()
        bar_json = self.get_job_json()
        bar_json['config']['requires'] = 'foo'
        rb_data['bar'] = bar_json
        lbs_data = set()
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, lbs_data, {})
        self.assertEqual([], mongo_data['tests']['bar']['blocked'])

        rb_data['foo']['builds']['1']['result'] = FAILURE
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, set(), {})
        self.assertEqual(['foo'], mongo_data['tests']['bar']['blocked'])

    def test_test_make_mongo_data_finished(self):
        rb_data = self.get_revision_build_data()
        lbs_data = set()
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, lbs_data, {})
        self.assertEqual('1970-01-01T00:00:00.300000Z', mongo_data['finished'])

        rb_data['foo'] = self.get_job_json(timestamp=400, duration=0)
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, lbs_data, {})
        self.assertEqual('1970-01-01T00:00:00.400000Z', mongo_data['finished'])

        rb_data['bar'] = self.get_job_json(timestamp=0, duration=500)
        mongo_data = OutcomeState.make_mongo_data(32, rb_data, lbs_data, {})
        self.assertEqual('1970-01-01T00:00:00.500000Z', mongo_data['finished'])

    def test_write_mongo_data(self):
        rb_data = self.get_revision_build_data()
        lbs_data = set()
        with self.make_outcome_state() as state, state:
            state.write_mongo_data(32, rb_data, lbs_data, {'aws': 'foo'})
            mongo_name = state.outcome_filename(state.mongo_filename(32))
            with open(mongo_name) as mongo_file:
                mongo_data = json.load(mongo_file)
        self.assertEqual(self.get_expected_mongo_data(), mongo_data)

    def test_write_mongo_data_no_build_revision(self):
        rb_data = self.get_revision_build_data()
        rb_data['build-revision1'] = rb_data.pop('build-revision')
        lbs_data = set()
        with self.make_outcome_state() as state, state:
            state.write_mongo_data(32, rb_data, lbs_data, {})
            mongo_name = state.mongo_filename(32)
            self.assertFalse(os.path.lexists(mongo_name))
