from datetime import datetime
import os
from unittest import TestCase

from mock import patch

from cidirector.utility import (
    locked_open_rw,
    open_rw,
    parse_iso_date,
    S3Storage,
    temp_dir,
    )


__metaclass__ = type


class TestParseISODate(TestCase):

    def test_parse_iso_date_float(self):
        parsed = parse_iso_date('1970-01-01T00:02:03.5Z')
        self.assertEqual(parsed, datetime(1970, 1, 1, 0, 2, 3, 500000))

    def test_parse_iso_date_int(self):
        parsed = parse_iso_date('1970-01-01T00:02:03Z')
        self.assertEqual(parsed, datetime(1970, 1, 1, 0, 2, 3))


class TestS3Storage(TestCase):

    def test_from_config(self):
        config = {'foo': 'bar', 's3_config_file': 'baz',
                  's3_archive_bucket_url': 'qux'}
        s3_storage = S3Storage.from_config(config)
        self.assertEqual(s3_storage.archive_bucket_url, 'qux')
        self.assertEqual(s3_storage.config_file, 'baz')

    def test_run_s3cmd_config(self):
        s3_storage = S3Storage('foo', None)
        with patch('subprocess.check_output') as mock_co:
            s3_storage.run_s3cmd(['bar', 'baz'])
        mock_co.assert_called_once_with(['s3cmd', '-c', 'foo', 'bar', 'baz'])

    def test_archive_version_url(self):
        s3_storage = S3Storage('foo', 's3://bucket')
        self.assertEqual(s3_storage.archive_version_url('2b'),
                         's3://bucket/juju-ci/products/version-2b')


class TestOpenRW(TestCase):

    def setUp(self):
        self.temp_dir_context = temp_dir()
        self.temp_dir = self.temp_dir_context.__enter__()
        self.filename = os.path.join(self.temp_dir, 'state-file')

    def tearDown(self):
        self.temp_dir_context.__exit__(None, None, None)

    def test_open_rw_creates(self):
        self.assertFalse(os.path.exists(self.filename))
        with open_rw(self.filename):
            pass
        self.assertTrue(os.path.exists(self.filename))

    def test_file_writable(self):
        with open_rw(self.filename) as test_file:
            test_file.write('foo')

    def test_file_readable(self):
        with open_rw(self.filename) as test_file:
            test_file.read()

    def test_no_truncate(self):
        with open(self.filename, 'w') as test_file:
            test_file.write('foo')
        with open_rw(self.filename) as test_file:
            self.assertEqual(test_file.read(), 'foo')


class TestLockedOpenRW(TestCase):

    def test_acquire_no_file(self):
        with temp_dir() as directory:
            filename = os.path.join(directory, 'state-file')
            with locked_open_rw(filename):
                self.assertTrue(os.path.exists(filename))

    def test_acquire_retains_contents(self):
        with temp_dir() as directory:
            filename = os.path.join(directory, 'state-file')
            with open(filename, 'w') as f:
                f.write('foo')
            with locked_open_rw(filename) as f:
                f.seek(0)
                self.assertEqual(f.read(), 'foo')
