#!/usr/bin/env python
from __future__ import print_function

from argparse import ArgumentParser
from datetime import (
    datetime,
    timedelta,
    )
import errno
import os
import logging
import re
from socket import setdefaulttimeout
import json

from cidirector import (
    BaseResultJudge,
    ConfigReader,
    get_build_parameters,
    make_jenkins,
    setup_logging,
    ResourcefulJob,
    )
from jobs import (
    BUILD_REVISION,
    FAILURE,
    parse_tags,
    SUCCESS,
)
from storage import (
    BUILDING,
    datetime_from_timestamp,
    FAILED,
    find_blockers,
    get_substrate_from_parsed,
    PENDING,
    SUCCEEDED,
    topo_sort,
    )
from utility import (
    date_str,
    ensure_dir,
    locked_open_rw,
    )


__metaclass__ = type


CI_DIRECTOR = 'ci-director'
CLOUD_HEALTH = 'cloud-health'


def parse_args():
    parser = ArgumentParser()
    parser.add_argument('-v', '--verbose', help='Verbose info',
                        action='store_true')
    parser.add_argument('--log-path', help='Where to write logs.',
                        default='ci-director.log')
    parser.add_argument('--log-count', help='The number of backups to keep.',
                        default=2, type=int)
    return parser.parse_args()


def get_cid_config(job_info, cloud_health=False):
    section = CLOUD_HEALTH if cloud_health else CI_DIRECTOR
    description = job_info.get('description')
    return ResourcefulJob.config_from_description(
        job_info['name'], description, logging.getLogger(), section=section)


def iter_job_builds(jenkins, last_completed):
    """Yield info on all ci-director jobs.

    jenkins is a jenkins.Jenkins instance.  last_completed is a dict of
    {job_name: last_completed_build} as emitted by scan_new_builds.  It
    represents the Jenkins state of the previous run, and is used to filter
    out builds that have already been seen.

    iterates through (job_name, unseen_builds, last_completed_build) tuples.
    """
    info = jenkins.get_info()
    for job in info['jobs']:
        job_name = job['name']
        logging.info('Scanning {}'.format(job_name))
        job_last = last_completed.get(job_name)
        job_info = jenkins.get_job_info(job_name)
        config = get_cid_config(job_info)
        if job_name != BUILD_REVISION and config is None:
            if get_cid_config(job_info, cloud_health=True) is None:
                logging.info('Skipping {}; no ci-director config'.format(
                             job_name))
                continue
        builds = sorted(b['number'] for b in job_info['builds'])
        builds = [b for b in builds if b > job_last]
        yield job_info, builds


def iter_build_info(jenkins, job_name, builds):
    for build in builds:
        logging.info('Scanning {} #{}'.format(job_name, build))
        yield jenkins.get_build_info(job_name, build)


def get_revision_build(job_name, build_info):
    """Determine the revision-build associated with this build_info.

    For build-revision, this is the build-revision's number.  For all other
    jobs, this is the build's revision_build parameter.
    """
    if job_name == BUILD_REVISION:
        return build_info['number']
    else:
        build_params = get_build_parameters(build_info)
        revision_build = build_params.get('revision_build', '')
        try:
            return int(revision_build)
        except ValueError:
            return None


def update_job_configs(jobs_by_name, builds_by_key):
    # Ensure config data is provided even for jobs with no builds.
    active_jobs = set(n for n, j in jobs_by_name.items()
                      if j['color'] != 'disabled')
    configs = {}
    active_jobs = set()
    for job_name, job_info in jobs_by_name.items():
        if job_info['color'] != 'disabled':
            active_jobs.add(job_name)
        config = get_cid_config(job_info, cloud_health=True)
        if config is None:
            config = get_cid_config(job_info, cloud_health=False)
        configs[job_name] = config

    for key_jobs in builds_by_key.values():
        # Include now-disabled jobs that have builds.
        for job_name in active_jobs.union(key_jobs.keys()):
            rb_job = key_jobs.setdefault(job_name, {'builds': {}})
            rb_job['config'] = configs[job_name]


def scan_new_builds(last_completed, jenkins):
    """Retrieve information on new builds from Jenkins.

    The last_completed parameter is a dict of job name to lastCompletedBuild
    number for the previous run, used to determine which builds are new.  Only
    builds with numbers greater than those in the last_completed parameter
    will be reported.

    jenkins is a jenkins.Jenkins instance.

    Returns a pair of dicts:
    by_revision_build is a dict of Jenkins build information, organized by
    revision-build, job name, build-number.  For example:
      {3096: {'my-job': {2850: {'number': 2850...}}}

    new_last_completed is in the same format as last_completed, but reflects
    the current Jenkins state.
    """
    by_revision_build = {}
    cloud_health = {}
    new_last_completed = {}
    revision_build_jobs = {}
    cloud_jobs = {}
    for job_info, builds in iter_job_builds(jenkins, last_completed):
        is_cloud = bool(get_cid_config(job_info, cloud_health=True) is not
                        None)
        job_name = job_info['name']
        if is_cloud:
            cloud_jobs[job_name] = job_info
        else:
            revision_build_jobs[job_name] = job_info
        last_completed_build = job_info['lastCompletedBuild']
        if last_completed_build is not None:
            new_last_completed[job_name] = last_completed_build['number']

        build_iter = iter_build_info(jenkins, job_name, builds)
        for build_info in build_iter:
            if is_cloud:
                date = datetime_from_timestamp(build_info['timestamp'])
                date_str = date.date().isoformat()
                by_job = cloud_health.setdefault(date_str, {})
            else:
                revision_build = get_revision_build(job_name, build_info)
                if revision_build is None:
                    continue
                by_job = by_revision_build.setdefault(revision_build, {})
            by_build_number = by_job.setdefault(job_name, {})
            builds = by_build_number.setdefault('builds', {})
            builds[str(build_info['number'])] = build_info
    update_job_configs(revision_build_jobs, by_revision_build)
    update_job_configs(cloud_jobs, cloud_health)
    return by_revision_build, cloud_health, new_last_completed


def json_dump_pretty(content, json_file):
    json.dump(content, json_file, indent=2, sort_keys=True,
              separators=(',', ': '))
    json_file.write('\n')


class OutcomeJob:
    """Job for use with BaseResultJudge and find_blockers."""

    def __init__(self, test_id, build_number, status, failure_count,
                 voting_build, tags, prerequisite_jobs, blocked):
        self.test_id = test_id
        self.build_number = build_number
        self.status = status
        self.failure_count = failure_count
        self.voting_build = voting_build
        self.tags = tags
        self.prerequisite_jobs = prerequisite_jobs
        self.blocked = blocked

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        for attribute in [
                'test_id', 'build_number', 'status', 'failure_count',
                'voting_build', 'tags', 'prerequisite_jobs', 'blocked']:
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        else:
            return True

    def __neq__(self, other):
        return not self.__eq__(other)

    @classmethod
    def from_json(cls, test_id, json, last_build_healthy):
        build_numbers = []
        building = False
        succeeded = False
        failure_count = 0
        if test_id == 'build-revision':
            config = {
                'tags': ('subjects=release function=make-tarball arch=all'
                         ' series=all substrate=none'),
            }
        else:
            config = json['config']
        failure_threshold = config.get('failure_threshold', 1)
        voting_build = config.get('vote', True) and last_build_healthy
        tags = config.get('tags')
        prerequisite_jobs = set(config.get('requires', '').split())
        for build in json['builds'].values():
            build_numbers.append(build['number'])
            if build['building']:
                building = True
            if build.get('result') == SUCCESS:
                succeeded = True
            if build.get('result') == FAILURE:
                failure_count += 1
        if succeeded:
            status = SUCCEEDED
        elif building:
            status = BUILDING
        elif failure_count >= failure_threshold:
            status = FAILED
        else:
            status = PENDING

        build_number = None if len(build_numbers) == 0 else max(build_numbers)
        return cls(test_id, build_number, status, failure_count, voting_build,
                   tags, prerequisite_jobs, blocked=[])

    def get_status(self):
        return self.status

    def is_blocked(self):
        return len(self.blocked) > 0


class OutcomeJobSource:
    """JobSource for use with BaseResultJudge."""

    @classmethod
    def from_build_data(cls, build_data, last_build_sickly):
        """Create OutcomeJobSource from build data.

        build_data is formatted as emitted by
        OutcomeState.update_revision_builds.
        """
        jobs = [OutcomeJob.from_json(j, b, j not in last_build_sickly)
                for j, b in build_data.items()]
        for job in topo_sort(jobs):
            job.blocked = find_blockers(job, jobs)
        return cls(jobs)

    def __init__(self, jobs):
        self.jobs = jobs

    def get_candidate_determine_result_jobs(self):
        return self.jobs


class NotLocked(Exception):
    """Raised when attempting to write to a directory that is not locked."""


class BuildFiles:
    """A Layer on top of OutcomeState for a set of build result files."""

    def __init__(self, outcome_state, filename_template):
        self.outcome_state = outcome_state
        self.filename_template = filename_template

    @classmethod
    def for_revision_builds(cls, state):
        return BuildFiles(state, 'revision-build-{}.json')

    @classmethod
    def for_cloud_health(cls, state):
        return BuildFiles(state, 'cloud-health-{}.json')

    def filename(self, file_key):
        return self.filename_template.format(file_key)

    @staticmethod
    def merge_revision_build_data(recorded_builds, new_builds):
        """Merge data from new_builds into recorded_builds.

        The job config data is updated only if new_builds has builds for that
        job.
        """
        for job, job_dict in new_builds.items():
            old_job_dict = recorded_builds.setdefault(
                job, {'builds': {}})
            old_job_dict['builds'].update(job_dict['builds'])
            if len(job_dict['builds']) > 0:
                if 'config' not in job_dict:
                    print('No config'.format(job_dict))
                    continue
                old_job_dict['config'] = job_dict['config']

    def load_json(self, file_key):
        filename = self.filename(file_key)
        return self.outcome_state.load_json(filename)

    def update_files(self, build_data):
        """Update the relevant files with new build data."""
        for revision_build, new_builds in build_data.items():
            recorded_builds = self.load_json(revision_build)
            if recorded_builds is None:
                recorded_builds = new_builds
            else:
                self.merge_revision_build_data(recorded_builds, new_builds)
            filename = self.filename(revision_build)
            self.outcome_state.write_json(filename, recorded_builds)
            yield revision_build, recorded_builds


def load_cloud_health(cloud_health_builds, build_days):
    result = {}
    for day in build_days:
        file_key = day.isoformat()
        result_json = cloud_health_builds.load_json(file_key)
        if result_json is None:
            continue
        result[file_key] = result_json
    return result


def get_substrate(config):
    parsed_tags = parse_tags(config['tags'])
    return get_substrate_from_parsed(parsed_tags)


def find_cloud_health_builds(cloud_health_builds, substrate, start, end):
    """Find cloud health builds matching a substrate and time window.

    cloud_health_builds is a dict of cloud_health_builds as emitted by
    load_cloud_health.

    substrate is the name of the substrate to match, as seen in the Jenkins
    ci-director config.

    start and end express the window as a pair of datetimes.  builds that
    ended before the start of the window or started after the end of the
    window are ignored.
    """
    for jobs in cloud_health_builds.values():
        for job_name, job in jobs.items():
            if get_substrate(job['config']) != substrate:
                continue
            for buildnum, build_info in job['builds'].items():
                duration = build_info.get('duration')
                # A build with no duration has no result, so it's not
                # interesting anyhow.
                if duration is None:
                    continue
                build_start = datetime_from_timestamp(build_info['timestamp'])
                build_end = datetime_from_timestamp(
                    build_info['timestamp'] + duration)
                if start > build_end:
                    continue
                if end < build_start:
                    continue
                yield job_name, build_info


def get_health_match_data(by_revision_build):
    """Return data on the last build for matching cloud health.

    Provides a dict with keys 'substrate', 'start', 'end'.  'start' and 'end'
    are extended by an hour to include cloud health builds that ended at most
    an hour before the revision build or started at most an hour after the
    revision build.

    Provides the days which may contain relevant cloud health as a set of
    datetime.date objects.
    """
    lb_data = {}
    lb_days = set()
    for revision_build, jobs in by_revision_build.items():
        rb_data = {}
        for job_name, job in jobs.items():
            if len(job['builds']) == 0:
                continue
            if job['config'] is None:
                continue
            last_build = sorted(
                job['builds'].values(), key=lambda b: b['number'])[-1]
            duration = last_build.get('duration')
            if duration is None:
                continue
            start = datetime_from_timestamp(last_build['timestamp'])
            end = start + timedelta(milliseconds=duration)
            start -= timedelta(hours=1)
            lb_days.add(start.date())
            lb_days.add(end.date())
            rb_data[job_name] = {
                'substrate': get_substrate(job['config']),
                'start': start,
                'end': end,
                }
        lb_data[revision_build] = rb_data
    return lb_data, lb_days


def find_last_build_sickly(health_match_data, cloud_health_build_dict):
    """Determine substrate health at build times.

    This uses the health match data to look up cloud health build results.

    It returns a tuple of last_build_sickly, substrates_health.

    last_build_sickly is a set of all jobs whose last build was done while its
    substrate was sickly.

    substrates_health is a list of all build results and their status at the
    time that any job's last build was done on it.
    """
    last_build_sickly = {}
    substrates_health = {}
    for revision_build, jobs in health_match_data.items():
        sickly_jobs = set()
        for job, lb_info in jobs.items():
            substrate = lb_info['substrate']
            ch_builds = find_cloud_health_builds(
                cloud_health_build_dict, substrate, lb_info['start'],
                lb_info['end'])
            results = set()
            for job_name, ch_build in ch_builds:
                result = ch_build.get('result')
                if result is None:
                    continue
                results.add(result)
                substrate_health = substrates_health.setdefault(
                    substrate, {})
                job_health = substrate_health.setdefault(job_name, {})
                job_health[str(ch_build['number'])] = {
                    'status': SUCCEEDED if result == SUCCESS else FAILED
                }
            results.discard(SUCCESS)
            if len(results) > 0:
                sickly_jobs.add(job)
        last_build_sickly[revision_build] = sickly_jobs
    return last_build_sickly, substrates_health


class OutcomeState:
    """Maintain information about job outcomes."""

    def __init__(self, outcome_dir):
        self.outcome_dir = outcome_dir
        self._lock_cxt = None
        self.last_completed_file = None
        self.data = None

    def outcome_filename(self, filename):
        """Return a filename relative to the outcome_dir."""
        return os.path.join(self.outcome_dir, filename)

    def __enter__(self):
        """Lock and open the last_completed file."""
        last_completed = self.outcome_filename('last-completed.json')
        ensure_dir(self.outcome_dir)
        self._lock_cxt = locked_open_rw(last_completed)
        self.last_completed_file = self._lock_cxt.__enter__()
        if os.fstat(self.last_completed_file.fileno()).st_size == 0:
            self.data = {}
        else:
            self.data = json.load(self.last_completed_file)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Write and unlock the last_completed file.

        Do not write on exception.
        """
        try:
            if exc_type is None:
                self.last_completed_file.seek(0)
                json_dump_pretty(self.data, self.last_completed_file)
                self.last_completed_file.truncate()
        finally:
            self.last_completed_file = None
            self._lock_cxt.__exit__(exc_type, exc_val, exc_tb)

    def load_json(self, relative_path):
        try:
            with open(self.outcome_filename(relative_path)) as build_file:
                    return json.load(build_file)
        except IOError as e:
            if e.errno != errno.ENOENT:
                raise
            return None

    def write_json(self, relative_path, content, make_dir=False):
        """Write a json file relative to the outcome_dir."""
        if self.last_completed_file is None:
            raise NotLocked('Write with no lock')
        filename = self.outcome_filename(relative_path)
        if make_dir:
            parent = os.path.dirname(filename)
            ensure_dir(parent)
        with open(filename, 'w') as json_file:
            json_dump_pretty(content, json_file)

    def mongo_filename(self, revision_build):
        return os.path.join(
            'version-{}'.format(revision_build), 'result.json')

    @staticmethod
    def make_mongo_data(revision_build, build_data, last_build_sickly,
                        substrates_health):
        """Make data formatted for juju-reports' mongodb.

        revision_build is the revision build number.
        build_data is a dict in the format produced by update_build_data.

        If build_data has no build-revision member, return None.

        last_build_sickly is a set of jobs whose last build was on a sickly
        substrate.

        substrates health is a list of health results for substrates at the
        time relevant builds were run.  i.e. it indicates whether the
        substrate was sickly at any point when tests were being run on it for
        this revision build.
        """
        if build_data.get('build-revision', {}).get('builds', {}) == {}:
            return None
        job_source = OutcomeJobSource.from_build_data(
            build_data, last_build_sickly)
        tests = {}
        for ojob in job_source.get_candidate_determine_result_jobs():
            tests[ojob.test_id] = {
                'name': ojob.test_id,
                'blocked': ojob.blocked,
                'failures': ojob.failure_count,
                'status': ojob.get_status(),
                'last_build': ojob.build_number,
                'vote': ojob.voting_build,
                'tags': ojob.tags,
            }
        build_revision_builds = build_data['build-revision']['builds']
        build_revision = build_revision_builds[str(revision_build)]
        release_number = None
        for artifact in build_revision['artifacts']:
            match = re.match(
                'juju-core_(\d+\.\d+.*).tar.gz', artifact['fileName'])
            if match:
                release_number = match.group(1)
        build_revision_params = get_build_parameters(build_revision)
        judge = BaseResultJudge(job_source)
        result = judge.determine_result()
        mongo_data = {
            'branch': build_revision_params['branch'],
            'build': {'last_build': revision_build},
            'complete_failure_info': dict(result.voting_failures),
            'cloud-health': {'substrates': substrates_health},
            'release_number': release_number,
            'result': {
                'status': result.result,
                },
            'revision_id': build_revision_params['revision'],
            'tests': tests,
        }

        def iter_updated():
            for job in build_data.values():
                for build in job['builds'].values():
                    yield build['timestamp'] + build['duration']

        last_updated_time = max(iter_updated()) / 1000.0
        last_updated = date_str(datetime.utcfromtimestamp(last_updated_time))
        # This varies from the original meaning; we treat "finished" as meaning
        # "last-updated".  This permits active revision-builds to be displayed
        # before their result is determined.
        mongo_data['finished'] = last_updated
        return mongo_data

    def write_mongo_data(self, revision_build, build_data, last_build_sickly,
                         substrates_health):
        """Write mongo data to disk.

        The data is formatted as emitted by make_mongo_data.

        revision_build is the revision build number.
        build_data is a dict in the format produced by update_build_data.

        If build_data has no build-revision member, do not do anything.

        last_build_sickly is a set of jobs whose last build was on a sickly
        substrate.

        substrates health is a list of health results for substrates at the
        time relevant builds were run.  i.e. it indicates whether the
        substrate was sickly at any point when tests were being run on it for
        this revision build.
        """
        mongo_data = self.make_mongo_data(revision_build, build_data,
                                          last_build_sickly,
                                          substrates_health)
        if mongo_data is None:
            return
        mongo_name = self.mongo_filename(revision_build)
        self.write_json(mongo_name, mongo_data, make_dir=True)


def main():
    """Entry point for update-outcome script."""
    args = parse_args()
    setup_logging(args.log_path, args.log_count, args.verbose)
    config = ConfigReader.read_config()
    setdefaulttimeout(30)
    jenkins = make_jenkins(config)
    outcome_dir = os.path.join(os.environ['HOME'],
                               '.config/ci-director-outcome')
    with OutcomeState(outcome_dir) as state:
        by_revision_build, cloud_health, new_last_completed = scan_new_builds(
            state.data, jenkins)
        cloud_health_builds = BuildFiles.for_cloud_health(state)
        list(cloud_health_builds.update_files(cloud_health))
        revision_build_files = BuildFiles.for_revision_builds(state)
        by_all_revision_build = dict(
            revision_build_files.update_files(by_revision_build))
        health_match_data, build_days = get_health_match_data(
            by_all_revision_build)
        cloud_health_build_dict = load_cloud_health(cloud_health_builds,
                                                    build_days)
        last_build_sickly, substrates_health = find_last_build_sickly(
            health_match_data, cloud_health_build_dict)
        for revision_build, build_data in by_all_revision_build.items():
            state.write_mongo_data(
                revision_build, build_data, last_build_sickly[revision_build],
                substrates_health)
        state.data.update(new_last_completed)


if __name__ == '__main__':
    main()
