from contextlib import contextmanager
from datetime import datetime
import errno
import fcntl
import os
from shutil import rmtree
import subprocess
from tempfile import mkdtemp


__metaclass__ = type


ISO_8601_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'


@contextmanager
def temp_dir():
    dirname = mkdtemp()
    try:
        yield dirname
    finally:
        rmtree(dirname)


def ensure_dir(dirpath):
    """Creates directories for dirpath if they don't already exist."""
    try:
        os.makedirs(dirpath)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


class S3Storage:
    """Object used to manipulate s3 storage.

    Holds configuration of credentials and s3 bucket.
    """
    def __init__(self, config_file, archive_bucket_url):
        self.config_file = config_file
        self.archive_bucket_url = archive_bucket_url

    @classmethod
    def from_config(cls, config):
        return cls(config['s3_config_file'], config['s3_archive_bucket_url'])

    def run_s3cmd(self, params):
        """Invoke s3cmd with specified config file."""
        # "s3cmd put ..." is chatty: It will always print something like
        # "file 'foo' stored as 's3://juju-qa-data/...'". There seems not to
        # any command line option like "--quiet", so let's catch and drop
        # these message here.
        subprocess.check_output(['s3cmd', '-c', self.config_file] + params)

    def archive_version_url(self, version):
        """Return first part of the S3 URL used to store CI test results."""
        return '%s/juju-ci/products/version-%s' % (
            self.archive_bucket_url, version)


def date_str(timestamp):
    return timestamp.isoformat() + 'Z'


def parse_iso_date(date_str):
    try:
        # Milliseconds must have six decimal places to be parsed correctly.
        # .5 will be handled as 5 milliseconds, whereas .500000 will be
        # handled as 500000 milliseconds.
        return datetime.strptime(date_str, ISO_8601_FORMAT)
    except ValueError:
        return datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%SZ')


def open_rw(filename):
    """Open a file for reading and writing.

    The file is created if it does not exist (like w+), but is not truncated
    (like r+).
    """
    fd = os.open(filename, os.O_RDWR | os.O_CREAT)
    return os.fdopen(fd, 'w+')


class FileInUse(Exception):
    """Raised when the lock for the state file is held by another process."""


@contextmanager
def locked_open_rw(file_name):
    """Open a file for read-write with an exclusive lock."""
    file_obj = open_rw(file_name)
    with file_obj:
        try:
            fcntl.lockf(file_obj, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError as e:
            if e.errno != errno.EAGAIN:
                raise
            raise FileInUse('File in use.')
        try:
            yield file_obj
        finally:
            file_obj.flush()
            fcntl.lockf(file_obj, fcntl.LOCK_UN)
